#include <cmath>
#include "cholroll.h"

using namespace std;

/*
 * MATLAB interface
 */

#include "mex.h"

/*
 * Main function: the inputs passed as "prhs[]" are:
 *
 * - L:     an N x N upper triangular matrix;
 * - new_M: a N x 1 double vector with the new row/column;
 */

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{    
    //Verify inputs: level 1
    if (nlhs != 0)
        mexErrMsgTxt("Required: no output arguments!");
    if (nrhs != 2 && nrhs != 4)
        mexErrMsgTxt("Required: two or four input arguments!");

    //Extract inputs
    
    const mxArray *xData0 = prhs[0];
    int M = mxGetN(xData0); //Number of row/columns
    double *L = mxGetPr(xData0);

    const mxArray *xData1 = prhs[1];
    double *new_M = mxGetPr(xData1); //New row/column.
    
    const mxArray *xData2; double *X = NULL;
    const mxArray *xData3; double *new_X = NULL;
    int   N;
    if (nrhs == 4) {
        xData2 = prhs[2];
        X      = mxGetPr(xData2);
        N      = mxGetN(xData2); //Number of columns
        xData3 = prhs[3];
        new_X  = mxGetPr(xData3);
    }

    do_update(L, new_M, M, X, new_X, N);
    
    return;
}

/*
 * Do in-place update of the upper triangular Cholesky factor "L" of a 
 * "M" x "M" matrix K, when the last row/column of K is changed. The new
 * row/column is passed as vector "new_M". Moreover, update "M" x "N" matrix 
 * X = L^{-T} * V, where V is also "M" x "N". In this case, we changed the
 * last row of V to "new_X". Do in-place update of "X".
 */

void do_update(double *L, double *new_M, int M, double *X, double *new_X, int N)
{
    int M1 = M - 1;

    /* Traverse first M-1 rows of L. L[M1 * M + i] gives L(i, M). */
    for (int i = 0; i < M1; i++) {
        double sum = new_M[i];
        for (int k = 0; k < i; k++)
            sum -= L[M1 * M + k] * L[i * M + k];
        L[M1 * M + i] = sum / L[i * M + i];
    }
    /* L(M, M) */
    double sum = new_M[M1];
    for (int k = 0; k < M1; k++)
        sum -= L[M1 * M + k] * L[M1 * M + k];
    L[M1 * M + M1] = sqrt(sum);

    /* Update X */
    if (X != NULL) {
        for (int j = 0; j < N; j++) { /* Go through columns j = 1, ..., N */
             X[j * M + M1] = new_X[j]; /* X(M, j) */
             for (int i = 0; i < M1; i++) /* Go through rows i = 1, ..., M - 1 */
                 X[j * M + M1] -= L[M1 * M + i] * X[j * M + i]; /* L transpose, so go through the Mth column of L */
             X[j * M + M1] /= L[M1 * M + M1];
        }
    }
}


