% mcmc_gpsem_full:
%
% Mind-boggling slow version of GPSEM which uses no pseudo-inputs. For
% small datasets (hundreds of points) or those that can't be convinced that
% full GPs are too slow for this.
%
% IT IS ASSUMED THAT ORDER OF OBSERVED VARIABLES IN THE GRAPH CORRESPONDS
% TO THE ORDER OF COLUMNS IN THE DATASET.
%
% IT IS ASSUMED THAT ORDER OF OBSERVED VARIABLES IN THE GRAPH CORRESPONDS
% TO THE ORDER OF COLUMNS IN THE DATASET.
%
% Inputs:
%
% - n_mcmc_samples: number of MCMC samples to be generated.
% - graph: a valid SEM DAG;
% - n_types: a binary vector indicating which variables in the graph
%            are observed (== 1) and which are latent (== 0). 
%            IT IS ASSUMED THAT ORDER OF OBSERVED VARIABLES IN THE GRAPH 
%            CORRESPONDS TO THE ORDER OF COLUMNS IN THE DATASET;
% - data: a N x D matrix, N being the sample size, D the number of
%         observable variables;
%
% - covfunc: string indicating the type of kernel used in the GP;
%
% - hyper_lin_var: global prior variance hyperparameter for the linear
%                  coefficients;
% - hyper_x_var:   inverse gamma hyperparameters for conditional latent
%                  variance (for endogenous latent variables)
% - hyper_x_mix:   parameter for the Dirichlet mixture model (for exogenous 
%                  latent variables)
% - hyper_y_par:   hyperparameters for the distribution of the measurement 
%                  error terms. For instance, inverse gamma hyperparameters 
%                  for measurement variance (if measurement model is
%                  Gaussian).
%
% - fixed_hyper_kernel: vector of zeros and ones. Ones indicate which
%                       hyperparameters are fixed using the values at
%                       'hyper_kernel_values';
% - fixed_yerr: vector of zeros and ones. Ones indicate which
%               measurement error variances are fixed using the values at
%               'yerr'
% - fixed_xerr: vector of zeros and ones. Ones indicate which
%               conditional latent variances are fixed using the values at
%               'xerr'
% - fixed_loadings: cell array, one entry per observed variable. Each
%                   entry is a vector indicating which loadings are fixed
%                   to '1' (the size of the vector is the number of parents
%                   of y plus one, where the last entry corresponds to the
%                   intercept).
%
% - missing_values: V x 2 matrix indicating the V variables in the data
%                   that should be treated as missing values. First element
%                   in each row indicates in which row the missing entry
%                   is. The second element indicates which column;
%
% - use_starting: see below;
% - starting_*: initial positions for latent variables, linear
%               coefficients, and error variances. Only used if 
%               'use_starting' is set to 1;
%
% Prior for kernel hyperparameters will vary according to the kernel.

function [sampled_missing, ...
          sampled_x, sampled_f, sampled_lin, sampled_kernel_hyper, ...
          sampled_xerr_var, sampled_xmix, sampled_x_z, sampled_yerr_par] = ...
          mcmc_gpsem_full(num_mcmc_samples, graph, n_types, data, covfunc, ...
                     hyper_lin_var, hyper_x_var, hyper_x_mix, hyper_y_par, ...                    
                     fixed_hyper_kernel, hyper_kernel_values, ...
                     fixed_yerr, yerr, fixed_xerr, xerr, ...
                     fixed_loadings, fixed_loading_values, ...
                     missing_values, ...
                     use_starting, ...
                     starting_latents, starting_loadings, starting_yvars, ...
                     starting_kernelhyper, starting_xvars, starting_xmix, ...
                     starting_exo_z, BATCH_SIZE)

% convert to cell if needed               
if ischar(covfunc), covfunc = cellstr(covfunc); end 
verbose = 0;

if BATCH_SIZE > num_mcmc_samples % Number of samples stored in primary memory
   BATCH_SIZE = num_mcmc_samples;
end
MH_X_PROP  = sqrt(0.6);  % Standard deviation of latent variable proposal

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PARSE GRAPH information.
%
% x_ychildren: observed children of each latent. It contains the position
%              of each children wrt to the observed variables only, not 
%              the whole graph.
% x_xchildren: latent children of each latent. Again, numbers represent
%              position with respect the latent variables only.
% y_xparents:  latent parents of each observed variable (same indexing
%              comment applies).

num_nodes = size(graph, 1);
latents = find(n_types == 0);  observed = find(n_types == 1);
num_latents = length(latents); num_observed = length(observed);
children = cell(num_nodes, 1); parents = cell(num_nodes, 1);
for v = 1:num_nodes
    children{v} = find(graph(:, v) == 1)';
    parents{v} = find(graph(v, :) == 1);
end

x_ychildren = cell(num_latents, 1); 
x_xchildren = cell(num_latents, 1); x_xparents = cell(num_latents, 1);
exo_latent = zeros(num_latents, 1);
for x = 1:num_latents
    x_pos = latents(x);
    obs_children    = intersect(observed, children{x_pos});
    latent_children = intersect(latents, children{x_pos});
    latent_parents  = intersect(latents, parents{x_pos});
    x_ychildren{x} = zeros(1, length(obs_children));
    for v = 1:length(obs_children)
        x_ychildren{x}(v) = find(observed == obs_children(v), 1);
    end
    x_xchildren{x} = zeros(1, length(latent_children));
    for v = 1:length(latent_children)
        x_xchildren{x}(v) = find(latents == latent_children(v), 1);
    end
    x_xparents{x} = zeros(1, length(latent_parents));
    for v = 1:length(latent_parents)
        x_xparents{x}(v) = find(latents == latent_parents(v), 1);
    end
    exo_latent(x) = isempty(latent_parents);
end
exo_latents = find(exo_latent == 1)'; endo_latents = find(exo_latent == 0)';
y_xparents = cell(1, num_observed);
for y = 1:num_observed
    y_pos = observed(y);
    y_xparents{y} = zeros(1, length(parents{y_pos}));
    for v = 1:length(parents{y_pos})
        y_xparents{y}(v) = find(latents == parents{y_pos}(v), 1);
    end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ALLOCATE space. Everything is being done in RAM in batches, so this can 
% quickly exhauste the memory. Set BATCH_SIZE accordingly.

N  = size(data, 1);
sampled_x    = zeros(BATCH_SIZE + 1, N, num_latents);
sampled_f    = zeros(BATCH_SIZE + 1, N, num_latents);
sampled_lin  = cell(num_observed, 1);
for v = 1:num_observed
    y = observed(v);
    sampled_lin{v} = zeros(BATCH_SIZE + 1, length(parents{y}) + 1);
end
sampled_xerr_var = zeros(BATCH_SIZE + 1, num_latents);
sampled_xmix     = cell(BATCH_SIZE + 1, length(exo_latents));
sampled_x_z      = zeros(BATCH_SIZE + 1, N, length(exo_latents));
sampled_yerr_par = zeros(BATCH_SIZE + 1, num_observed);

for x = exo_latents
    exo_pos = find(exo_latents == x, 1);
    num_k = length(hyper_x_mix{x, 1});
    for i = 1:(BATCH_SIZE + 1)
        sampled_xmix{i, exo_pos} = zeros(num_k, 3);
    end
end

logtheta = cell(num_latents, 1);

% Cache all kernel matrices Cholesky factors (upper triangular ones)
L_M = cell(num_latents, 1);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% INITIAL SAMPLING. First, initialize error variances.
% After that, order latents in a partial order where a node 
% cannot come before any of its descendants. Then sample according to
% order.

if use_starting
   sampled_xerr_var(1, :) = starting_xvars;
   sampled_yerr_par(1, :) = starting_yvars;
   for d = 1:N
      sampled_x(1, d, :) = starting_latents(d, :);
   end
   for y = 1:num_observed
       sampled_lin{y}(1, :) = starting_loadings{y};
   end    
   sampled_kernel_hyper = cell(num_latents, 1);
   for x = 1:num_latents
       if ~isempty(x_xparents{x})
          sampled_kernel_hyper{x} = zeros(BATCH_SIZE + 1, ...
                           kernel_nparam(covfunc, length(x_xparents{x})));
          sampled_kernel_hyper{x}(1, :) = starting_kernelhyper{x}';
       end
   end
   for x = exo_latents
       exo_pos = find(exo_latents == x, 1);
       sampled_xmix{1, exo_pos} = starting_xmix{x};
       sampled_x_z(1, :, exo_pos) = starting_exo_z(x, :);
   end
   latent_order = intersect(get_node_order(graph), latents);
   for x_ord = 1:num_latents
       x = latent_order(x_ord);
       if isempty(x_xparents{x})
          continue
       end
       % Sample latent functions
       logtheta{x} = kernel_gpml_map(covfunc, sampled_kernel_hyper{x}(1, :), 1);
       if length(x_xparents{x}) == 1
          x_data = sampled_x(1, :, x_xparents{x})';
       else
          x_data = squeeze(sampled_x(1, :, x_xparents{x}));
       end
       K_M = feval(covfunc{:}, logtheta{x}, x_data);  
       L_M{x} = chol(K_M);
       sampled_f(1, :, x) = randnorm(1, zeros(N, 1), L_M{x});
   end
else
   sampled_kernel_hyper = cell(num_latents, 1);
   for x = 1:num_latents
       if ~isempty(x_xparents{x})
          sampled_kernel_hyper{x} = ones(BATCH_SIZE + 1, ...
                                       kernel_nparam(covfunc, length(x_xparents{x})));
          sampled_kernel_hyper{x}(:, fixed_hyper_kernel == 1) = ...
                                       repmat(hyper_kernel_values(fixed_hyper_kernel == 1)', BATCH_SIZE + 1, 1);
       end
   end
   sampled_xerr_var(1, :) = xerr';
   sampled_yerr_par(1, :) = yerr';
   latent_order = intersect(get_node_order(graph), latents);
   for x_ord = 1:num_latents
       x = latent_order(x_ord);
       if isempty(x_xparents{x})
          sampled_x(1, :, x) = randnorm(N, 1) * ...
                               sqrt(sampled_xerr_var(1, x));
          continue
       end
       % Sample latent functions 
       logtheta{x} = kernel_gpml_map(covfunc, sampled_kernel_hyper{x}(1, :), 1);
       if length(x_xparents{x}) == 1
          x_data = sampled_x(1, :, x_xparents{x})';
       else
          x_data = squeeze(sampled_x(1, :, x_xparents{x}));
       end
       K_M = feval(covfunc{:}, logtheta{x}, x_data);  
       L_M{x} = chol(K_M);
       sampled_f(1, :, x) = randnorm(1, zeros(N, 1), L_M{x});
       sampled_x(1, :, x) = sampled_f(1, :, x) + randn(1, N) * sqrt(sampled_xerr_var(1, x));
   end
   for y = 1:num_observed % And now, sample linear coefficients
       n_par = length(y_xparents{y});
       cov_y = eye(n_par + 1);
       sampled_lin{y}(1, :) = randnorm(1, zeros(n_par + 1, 1), [], cov_y)';
   end
   for x = exo_latents
       exo_pos = find(exo_latents == x, 1);
       num_k = length(hyper_x_mix{x, 1});
       sampled_x_z(1, :, exo_pos) = sample(ones(num_k, 1) / num_k, N);
       for k = 1:num_k
           sampled_xmix{1, exo_pos}(k, 1) = 1 / num_k; % Probability 
           sampled_xmix{1, exo_pos}(k, 2) = randn();   % Mean
           sampled_xmix{1, exo_pos}(k, 3) = rand();    % Variance
       end
   end
end

for y = 1:num_observed
    n_par = length(y_xparents{y});
    sampled_lin{y}(1, fixed_loadings{y} == 1) = 1;
    if fixed_loadings{y}(n_par + 1)
       sampled_lin{y}(1, n_par + 1) = 0;
    end
end

% Initialize missing values
num_missing = size(missing_values, 1);
count_vars = N * ones(num_observed, 1);
mars_matrix = zeros(size(data));

for t = 1:num_missing
    r = missing_values(t, 1); c = missing_values(t, 2);
    data(r, c) = 0;
    count_vars(c) = count_vars(c) - 1;
end
data_means = sum(data) ./ count_vars';
if num_missing > 0
   sampled_missing = zeros(BATCH_SIZE + 1, num_missing);
else
   sampled_missing = [];
end
for t = 1:num_missing
    r = missing_values(t, 1); c = missing_values(t, 2);
    data(r, c) = data_means(c) + 0.01 * randn();
    sampled_missing(1, t) = data(r, c);
    mars_matrix(r, c) = 1;
end

% Buffers
L_sub = cell(num_latents, 1);
cond_data = cell(num_latents, 1);

disp(sampled_yerr_par(1, :));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SETUP BATCH SIZE INFORMATION

num_batches = ceil(num_mcmc_samples / BATCH_SIZE);
num_batch_samples = BATCH_SIZE * ones(num_batches, 1);
num_batch_samples(num_batches) = num_mcmc_samples - BATCH_SIZE * (num_batches - 1);

file_x       = fopen('sample_x.txt','w');
file_f       = fopen('sample_f.txt','w');
file_exo     = fopen('sample_exo.txt', 'w');
file_xvar    = fopen('sample_xvar.txt', 'w');
file_y       = fopen('sample_y.txt','w');
file_z       = fopen('sample_z.txt','w');
file_hgp     = fopen('sample_hgp.txt','w');
file_lambda  = fopen('sample_lambda.txt','w');
file_yvar    = fopen('sample_yvar.txt','w');

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PERFORM SAMPLING

total_iter = 0;

for bt = 1:num_batches
for iter = 2:(num_batch_samples(bt) + 1)

    total_iter = total_iter + 1; disp(sprintf('Iteration %d', total_iter));

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Pre-arrangements
    
    sampled_x(iter, :, :) = sampled_x(iter - 1, :, :);
    sampled_xerr_var(iter, :) = sampled_xerr_var(iter - 1, :);
    for x = 1:num_latents
        if ~isempty(x_xparents{x})
           sampled_kernel_hyper{x}(iter, :) = sampled_kernel_hyper{x}(iter - 1, :);
           sampled_f(iter, :, x)            = sampled_f(iter - 1, :, x);
        end        
    end
    sampled_yerr_par(iter, :) = sampled_yerr_par(iter - 1, :);
    for y = 1:num_observed        
        sampled_lin{y}(iter, :) = sampled_lin{y}(iter - 1, :);
    end
        
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Sample latent functions given latent variables and data

    for x = 1:num_latents
        if ~isempty(parents{x})
            L_inv    = inv(L_M{x});
            K_inv    = L_inv * L_inv';
            K_pos    = inv(K_inv + eye(N) / sampled_xerr_var(iter, x));
            mean_pos = K_pos * sampled_x(iter, :, x)' / sampled_xerr_var(iter, x);
            chol_pos = chol(K_pos);
            sampled_f(iter, :, x) = randnorm(1, mean_pos, chol_pos);
        end
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Sample kernel hyperparameters

    for x = 1:num_latents
        if ~isempty(parents{x})
           if length(x_xparents{x}) == 1
              x_data = sampled_x(iter, :, x_xparents{x})';
           else
              x_data = squeeze(sampled_x(iter, :, x_xparents{x}));
           end
           np = kernel_nparam(covfunc, length(x_xparents{x}));           
           k_format0 = ' %.4f'; k_format = k_format0; for kf = 2:np, k_format = strcat(k_format, k_format0); end

           % Propose hyperparameter values
           curr_par = sampled_kernel_hyper{x}(iter, :);
           new_par = curr_par;
           logfwd = 0; logbkw = 0; log_currprior = 0; log_newprior = 0;
           for p = 1:np
               if fixed_hyper_kernel(p)
                  continue
               end
               [new_par_p logfwd_p logbkw_p] = kernel_prop(covfunc, p, curr_par(p));
               new_par(p) = new_par_p;
               logfwd = logfwd + logfwd_p;
               logbkw = logbkw + logbkw_p;
               log_newprior  = log_newprior  + kernel_logprior(covfunc, p, new_par(p));
               log_currprior = log_currprior + kernel_logprior(covfunc, p, curr_par(p));
           end
           old_logtheta = logtheta{x};
           logtheta{x} = kernel_gpml_map(covfunc, new_par, 1);
           newK_M = feval(covfunc{:}, logtheta{x}, x_data);  
           newL_M = chol(newK_M);
           f_data = sampled_f(iter, :, x)';
           new_f_norm = newL_M' \ f_data; new_f_norm = sum(new_f_norm .* new_f_norm);
           old_f_norm = L_M{x}' \ f_data; old_f_norm = sum(old_f_norm .* old_f_norm);
           new_logfactor = -sum(log(diag(newL_M))) - 0.5 * new_f_norm + ... 
                           log_newprior;
           old_logfactor = -sum(log(diag(L_M{x}))) - 0.5 * old_f_norm + ...
                           log_currprior;
           
           ratio = exp(new_logfactor + logbkw - old_logfactor - logfwd);

           if rand() < ratio % Accept it 
              L_M{x} = newL_M;
              sampled_kernel_hyper{x}(iter, :) = new_par;              
           else
              logtheta{x} = old_logtheta;
           end                       
           disp(sprintf(strcat(k_format), sampled_kernel_hyper{x}(iter, :), x));
        end
    end

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Sample latent variable conditional variance hyperparameters
    % (for endogenous latent variables)
    
    for x = endo_latents
        if fixed_xerr(x)
           continue
        end
        ss_x = sum((sampled_x(iter, :, x) - sampled_f(iter, :, x)) .^ 2);
        ga = hyper_x_var(1) + N / 2;
        gb = hyper_x_var(2) + ss_x / 2;
        sampled_xerr_var(iter, x) = gb / randgamma(ga);
    end

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Sample latent variable marginal (mixture of Gaussians) parameters
    % (for exogenous latent variables)
    
    for x = exo_latents        
        exo_pos = find(exo_latents == x, 1);
        num_k = length(hyper_x_mix{x, 1});
        % First, sample means and variances for each mixture component        
        for k = 1:num_k
            k_rows = find(sampled_x_z(iter - 1, :, exo_pos) == k);
            s_x = sum(sampled_x(iter - 1, k_rows, x));            
            % Sample mean give variance
            v_mean = 1 / (1 / hyper_x_mix{x, 2} + ...
                     length(k_rows) / sampled_xmix{iter - 1, exo_pos}(k, 3));
            m_mean = v_mean * s_x / sampled_xmix{iter - 1, exo_pos}(k, 3);
            new_mean = randn() * sqrt(v_mean) + m_mean;
            sampled_xmix{iter, exo_pos}(k, 2) = new_mean;
            % Sample variance given mean
            ss_x = sum((sampled_x(iter - 1, k_rows, x) - new_mean).^2);
            ga = hyper_x_mix{x, 3}(1) + length(k_rows) / 2;
            gb = hyper_x_mix{x, 3}(2) + ss_x / 2;
            sampled_xmix{iter, exo_pos}(k, 3) = gb / randgamma(ga);
        end
        % Now sample mixture component indicators given x and means
        for d = 1:N
            weights = zeros(num_k, 1);
            for k = 1:num_k
                m = sampled_xmix{iter, exo_pos}(k, 2);
                v = sampled_xmix{iter, exo_pos}(k, 3);
                xd = sampled_x(iter - 1, d, x);
                weights(k) = -0.5 * log(v) - 0.5 * (xd - m)^2 / v + ...
                             log(sampled_xmix{iter - 1, exo_pos}(k, 1));
            end
            weights = weights - max(weights); weights = exp(weights);
            sampled_x_z(iter, d, exo_pos) = sample(weights/sum(weights));
        end
        % Finally, sample mixture distribution
        alpha = hyper_x_mix{x, 1};
        for k = 1:num_k
            alpha(k) = alpha(k) + sum(sampled_x_z(iter, :, exo_pos) == k);
        end
        new_prob = randdirichlet(alpha);
        for k = 1:num_k
            sampled_xmix{iter, exo_pos}(k, 1) = new_prob(k);
        end        
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Sample measurement error hyperparameters
    
    for y = 1:num_observed
        if fixed_yerr(y)
           continue
        end
        recN  = mars_matrix(:, y) == 0;
        if length(y_xparents{y}) > 1
           x_data = [squeeze(sampled_x(iter, recN, y_xparents{y})) repmat(1, sum(recN), 1)];
        else
           x_data = [sampled_x(iter, recN, y_xparents{y})' repmat(1, sum(recN), 1)]; 
        end        
        ss_y = sum((data(recN, y) - x_data * sampled_lin{y}(iter, :)') .^ 2);
        ga = hyper_y_par(1) + sum(recN) / 2;
        gb = hyper_y_par(2) + ss_y / 2;
        sampled_yerr_par(iter, y) = gb / randgamma(ga);
    end
    disp(sampled_yerr_par(iter, :));
        
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Sample linear coefficients given latent variables and data

    for y = 1:num_observed        
        recN  = mars_matrix(:, y) == 0;
        n_par = length(y_xparents{y});
        if length(y_xparents{y}) > 1
           x_data = [squeeze(sampled_x(iter, recN, y_xparents{y})) repmat(1, sum(recN), 1)];
        else
           x_data = [sampled_x(iter, recN, y_xparents{y})' repmat(1, sum(recN), 1)];
        end
        cov_y = x_data' * x_data / sampled_yerr_par(iter, y);
        for i = 1:(n_par + 1)
            cov_y(i, i) = cov_y(i, i) + 1 / hyper_lin_var;
        end
        cov_y = inv(cov_y);
        mean_y = cov_y * sum(repmat(data(recN, y), 1, n_par + 1) .* x_data)' / ...
                         sampled_yerr_par(iter, y); 
        if sum(fixed_loadings{y} == 0)
           sampled_lin{y}(iter, :) = randnorm(1, mean_y, [], cov_y)';        
        else
           fr = fixed_loadings{y} == 0; % Free coefficients
           fx = fixed_loadings{y} == 1; % Fixed coefficients
           if ~isempty(fixed_loading_values)
              sampled_lin{y}(iter, fx) = fixed_loading_values{y}';
           else
              sampled_lin{y}(iter, fx) = 1;            
              if fixed_loadings{y}(n_par + 1)
                 sampled_lin{y}(iter, n_par + 1) = 0;
               end
           end            
           reg_coeff = cov_y(fr, fx) * inv(cov_y(fx, fx)); 
           mean_f = reg_coeff * (sampled_lin{y}(iter, fx)' - mean_y(fx));
           cov_f  = cov_y(fr, fr) - reg_coeff * cov_y(fx, fr);
           sampled_lin{y}(iter, fr) = randnorm(1, mean_f, [], cov_f)';
         end
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Sample latent variables
    
    sample_latents();

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Sample missing values
    
    if num_missing > 0
       sds = sqrt(sampled_yerr_par(iter, :));
       rand_pool = randn(num_missing, 1);
       for t = 1:num_missing
           d = missing_values(t, 1);
           y = missing_values(t, 2);        
           m_t = sampled_lin{y}(iter, :) * [sampled_x(iter, d, y_xparents{y}); 1];
           data(d, y) = m_t + sds(y) * rand_pool(t);
           sampled_missing(iter, t) = m_t; % Report mean only
       end
    end

    %sample_pseudo_x_cascade();
    if verbose > 4
       print_coeff();
    end
    
end % batch

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% NEXT BATCH preparation

last_s = num_batch_samples(bt) + 1;
print_to_file(last_s); % Print whole batch to file
sampled_x(1, :, :) = sampled_x(last_s, :, :);
sampled_f(1, :, :) = sampled_f(last_s, :, :);
for v = 1:num_observed
    sampled_lin{v}(1, :) = sampled_lin{v}(last_s, :);
end
sampled_xerr_var(1, :) = sampled_xerr_var(last_s, :);
for x = exo_latents
    exo_pos = find(exo_latents == x, 1);
    sampled_xmix{1, exo_pos} = sampled_xmix{last_s, exo_pos};
end
sampled_x_z(1, :, :)   = sampled_x_z(last_s, :, :);
sampled_yerr_par(1, :) = sampled_yerr_par(last_s, :);

for x = 1:num_latents
    if ~isempty(parents{x})
       sampled_kernel_hyper{x}(1, :) = sampled_kernel_hyper{x}(last_s, :);
    end
end
if num_missing > 0
   sampled_missing(1, :) = sampled_missing(last_s, :);
end

end % all batches


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CLEAN-UP

fclose(file_x); fclose(file_f); fclose(file_exo); fclose(file_xvar);
fclose(file_y); fclose(file_z); fclose(file_hgp);
fclose(file_lambda); fclose(file_yvar); 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Local SAMPLING functions

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Hard part: sample latent variables given latent functions, 
% linear coefficients and data    

function sample_latents()
    if verbose > 0; disp('Sampling latents... '); end
    num_accepted = zeros(num_latents, 1);
    
    for xx = 1:num_latents        
    
        perm_N = randperm(N);
        
        for xn = perm_N %#ok<FXUP>
            
            if verbose > 1, disp(sprintf('Sampled latent %d', xm)); end
            other_n = [1:(xn - 1) (xn + 1):N];
            
            %%%%% Cache children Choleskys after row/column removal and
            %%%%% data 
            for x_c = x_xchildren{xx}
                L_sub{x_c} = subchol(L_M{x_c}, xn);
                %K_M = L_M{x_c}' * L_M{x_c}; L_sub2 = chol(K_M(other_n, other_n)); disp(max(abs(L_sub{x_c} - L_sub2));
                if length(x_xparents{x_c}) == 1
                   cond_data{x_c} = sampled_x(iter, other_n, x_xparents{x_c})';
                else
                   cond_data{x_c} = squeeze(sampled_x(iter, other_n, x_xparents{x_c}));
                end               
            end

            %%%%%% Proposed and collect factors for new and old
            %%%%%% latent variable (xx, xn)
            logfactor1 = latent_logfactors(xn, xx, other_n);
            old_sampled_x = sampled_x(iter, xn, xx);
            sampled_x(iter, xn, xx) = old_sampled_x + MH_X_PROP * randn();
            logfactor2 = latent_logfactors(xn, xx, other_n);
                    
            % Evaluate proposal
            ratio = exp(logfactor2 - logfactor1);
            if rand() > ratio % Reject it
               sampled_x(iter, xn, xx) = old_sampled_x;
            else
               % Re-calculate Cholesky factors
               for x_c = x_xchildren{xx}
                   pred_data = squeeze(sampled_x(iter, xn, x_xparents{x_c}))';
                   [kk_star k_star] = feval(covfunc{:}, logtheta{x_c}, cond_data{x_c}, pred_data);                   
                   L_M{x_c}(1:(N - 1), 1:(N - 1)) = L_sub{x_c};
                   cholroll(L_M{x_c}, [k_star' kk_star]); % Update Cholesky to include new row/column
                   %L_M{x_c} = slow_cholroll(L_M{x_c}, [k_star' kk_star]);
                   if xn < N
                      cholupexch(L_M{x_c}, xn, N, 1); % Reorganize Cholesky to reflect original order
                   end
                   %K_temp = feval(covfunc{:}, logtheta{x_c}, sampled_x(iter, :, x_xparents{x_c})'); L_temp = chol(K_temp); max(max(abs(L_temp - L_M{x_c})))
               end
               num_accepted(xx) = num_accepted(xx) + 1;
            end

            for x_c = x_xchildren{x}
                L_sub{x_c} = [];
                cond_data{x_c} = [];
            end
            
        end

    end
    
    disp(num_accepted');
    if verbose > 0, disp('Sampling latents... DONE'); end
end

%
% Return the log-factors corresponding to the conditional likelihood of a
% latent variable given everybody else, as required for the
% Metropolis-Hastings update
%
% "xn" is the row number and "xx" the variable number.
%

function logfactor = latent_logfactors(xn, xx, other_n)
    % Factor due to latent parent function/marginal
    if ~exo_latent(xx)
       logfactor = -0.5 * (sampled_x(iter, xn, xx) - ...
                           sampled_f(iter, xn, xx))^2 / sampled_xerr_var(iter, xx);
    else       
       exo_pos = find(exo_latents == xx, 1);
       zz  = sampled_x_z(iter, xn, exo_pos);
       mms = sampled_xmix{iter, exo_pos}(zz, 2);
       vrs = sampled_xmix{iter, exo_pos}(zz, 3);
       logfactor = -0.5 * (sampled_x(iter, xn, xx) - mms)^2 / vrs;        
    end
    % Factors coming from each of its latent children
    for x_c = x_xchildren{xx}
        % Get predictive mean
        alpha  = solve_chol(L_sub{x_c}, sampled_f(iter, other_n, x_c)');
        pred_data = squeeze(sampled_x(iter, xn, x_xparents{x_c}))';
        [kk_star k_star] = feval(covfunc{:}, logtheta{x_c}, cond_data{x_c}, pred_data);
        pred_mean = k_star' * alpha;
        % Get predictive variance
        vv = L_sub{x_c}' \ k_star;
        pred_var = kk_star - vv' * vv;                
        % Add loglikelihood term
        logfactor = logfactor - 0.5 * (log(pred_var) + ...
                    (sampled_f(iter, xn, x_c) - pred_mean)^2 / pred_var);
    end
    % Factors coming from each of its observed children
    for y_c = x_ychildren{xx}
        if mars_matrix(xn, y_c) == 1
           continue
        end
        x_data = [squeeze(sampled_x(iter, xn, y_xparents{y_c}))' 1];
        b_y = sampled_lin{y_c}(iter, :)';
        logfactor = logfactor - ...
                    0.5 * (data(xn, y_c) - x_data * b_y)^2 / sampled_yerr_par(iter, y_c);
    end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Local AUXILIARY functions

% Print coefficients (for debugging purposes)

function print_coeff()
    for yy = 1:num_observed
        disp(sampled_lin{yy}(iter, :));
    end
end

% Print all points in the batch to the respective files

function print_to_file(endpoint)
   disp('Printing to file...');
   % Dump latent variables
   x_format = ' %.3f';
   x_format_r = x_format;
   for w = 1:(num_latents - 1)
       x_format_r = strcat(x_format_r, x_format);
   end
   x_format_r = strcat(x_format_r, '\n');
   for s = 2:endpoint
       fprintf(file_x, x_format_r, sampled_x(s, :, :));    
   end
   % Dump imputed data
   if num_missing > 0
      y_format = ' %.3f';
      y_format_r = y_format;
      for w = 1:(num_missing - 1)
          y_format_r = strcat(y_format_r, y_format);
      end
      y_format_r = strcat(y_format_r, '\n');
      for s = 2:endpoint
          fprintf(file_y, y_format_r, sampled_missing(s, :));    
      end
   end
   % Dump exogenous mixture indicators
   z_format = ' %d';
   z_format_r = z_format;
   for w = 2:length(exo_latents)
       z_format_r = strcat(z_format_r, z_format);
   end
   z_format_r = strcat(z_format_r, '\n');
   for s = 2:endpoint
       fprintf(file_z, z_format_r, sampled_x_z(s, :, :));    
   end
   % Dump GP hyperparameters
   hgp_format = ' %.3f';
   for s = 2:endpoint
       for cl = 1:num_latents
           if isempty(parents{cl})
              continue
           end
           hgp_format_r = hgp_format;
           for w = 1:(length(sampled_kernel_hyper{cl}(s, :)) - 1)
               hgp_format_r = strcat(hgp_format_r, hgp_format);
           end
           fprintf(file_hgp, hgp_format_r, sampled_kernel_hyper{cl}(s, :));    
       end
       fprintf(file_hgp, '\n');    
   end   
   % Dump function values
   f_format_r = x_format_r;
   for s = 2:endpoint
       fprintf(file_f, f_format_r, sampled_f(s, :, :));    
   end
   % Dump latent variances
   xvar_format = ' %.3f';
   for s = 2:endpoint
       for xx = 1:num_latents
           fprintf(file_xvar, xvar_format, sampled_xerr_var(s, xx));    
       end
       fprintf(file_xvar, '\n');    
   end   
   % Dump exogenous information
   exo_format = ' %.3f';
   for s = 2:endpoint
       for xx = 1:num_latents
           if ~isempty(parents{xx})
              continue
           end
           exo_pos = find(exo_latents == xx, 1);
           num_k = length(hyper_x_mix{xx, 1});
           for zz = 1:num_k
               fprintf(file_exo, exo_format, sampled_xmix{s, exo_pos}(zz, 1)); 
               fprintf(file_exo, exo_format, sampled_xmix{s, exo_pos}(zz, 2)); 
               fprintf(file_exo, exo_format, sampled_xmix{s, exo_pos}(zz, 3)); 
           end
       end
       fprintf(file_exo, '\n');    
   end   
   % Dump linear coefficient hyperparameters
   lambda_format = ' %.3f';
   for s = 2:endpoint
       for yy = 1:num_observed
           lambda_format_r = lambda_format;
           for w = 1:length(y_xparents{yy})
               lambda_format_r = strcat(lambda_format_r, lambda_format);
           end
           fprintf(file_lambda, lambda_format_r, sampled_lin{yy}(s, :));    
       end
       fprintf(file_lambda, '\n');    
   end   
   % Dump variances
   var_format = ' %.3f';
   for s = 2:endpoint
       for yy = 1:num_observed
           fprintf(file_yvar, var_format, sampled_yerr_par(s, yy));    
       end
       fprintf(file_yvar, '\n');    
   end   
end



end

