% mcmc_gpsem_spgp:
%
% ``Fast'' version of GPSEM with pseudo-inputs. The GPSEM version with
% standard Gaussian processes (mcmc_gpsem_full) is essentially unworkable
% for more than a couple of hundreds of datapoints. The GP prior
% implemented here is a variation of the pseudoinputs models of Snelson and
% Ghahramani, where pseudoinputs are treated in a full Bayesian way.
%
% IT IS ASSUMED THAT ORDER OF OBSERVED VARIABLES IN THE GRAPH CORRESPONDS
% TO THE ORDER OF COLUMNS IN THE DATASET.
%
% Inputs:
%
% - n_mcmc_samples: number of MCMC samples to be generated;
% - M: size of pseudoinput layer;
% - graph: a valid SEM DAG;
% - n_types: a binary vector indicating which variables are observed
%            (== 1) and which are hidden (== 0)
% - data: a N x D matrix with data for the observed variables. IT IS 
%         ASSUMED THAT ORDER OF OBSERVED VARIABLES IN THE GRAPH CORRESPONDS
%         TO THE ORDER OF COLUMNS IN THE DATASET.
% - covfunc: string indicating the type of kernel used in the GP
% - lin_var: global prior variance hyperparameter for the linear
%            coefficients
% - hyper_x_var: inverse gamma hyperparameters for conditional latent
%                variance (for endogenous latent variables)
% - hyper_x_mix: parameters for the Dirichlet mixture model (for exogenous 
%                latent variables)
% - hyper_y_par: hyperparameters for the distribution of the measurement 
%                error terms. For instance, inverse gamma hyperparameters 
%                for measurement variance (if 'measurement_type' is
%                Gaussian).
% - fixed_hyper_kernel: vector of zeros and ones. Ones indicate which
%                       hyperparameters are fixed using the values at
%                       'hyper_kernel_values'
% - fixed_yerr: vector of zeros and ones. Ones indicate which
%               measurement error variances are fixed using the values at
%               'yerr'
% - fixed_xerr: vector of zeros and ones. Ones indicate which
%               conditional latent variances are fixed using the values at
%               'xerr'
% - fixed_loadings: cell array, one entry per observed variable. Each
%                   entry is a vector indicating which loadings are fixed
%                   to the values in 'fixed_loading_values' (the size of 
%                   the vector is the number of parents
%                   of y plus one, where the last entry corresponds to the
%                   intercept).
% - missing_values: V x 2 matrix indicating the V variables in the data
%                   that should be treated as missing values. First element
%                   in each row indicates in which row the missing entry
%                   is. The second element indicates which column;
% - use_starting: see below.
% - starting_*: initial positions for latent variables, linear
%               coefficients, and error variances. Only used if 
%               'use_starting' is set to 1.
%
% Prior for kernel hyperparameters will vary according to the kernel.
%
% IT IS ASSUMED THAT ORDER OF OBSERVED VARIABLES IN THE GRAPH CORRESPONDS
% TO THE ORDER OF COLUMNS IN THE DATASET.
%

function [sampled_missing, ...
          sampled_x, sampled_f, sampled_lin, sampled_kernel_hyper, ...
          sampled_xerr_var, sampled_xmix, sampled_x_z, sampled_yerr_par, pseudo_rho] = ...
          mcmc_gpsem_spgp(num_mcmc_samples, M, graph, n_types, data, covfunc, lin_var, ...
                     hyper_x_var, hyper_x_mix, hyper_y_par, ...                    
                     fixed_hyper_kernel, hyper_kernel_values, ...
                     fixed_yerr, yerr, fixed_xerr, xerr, ...
                     fixed_loadings, fixed_loading_values, ...
                     missing_values, ...
                     use_starting, ...
                     starting_latents, starting_loadings, starting_yvars, ...
                     starting_kernelhyper, starting_xvars, starting_xmix, ...
                     starting_exo_z, P_VARIANCE, BATCH_SIZE)

% convert to cell if needed               
if ischar(covfunc), covfunc = cellstr(covfunc); end 
verbose = 0;
if BATCH_SIZE > num_mcmc_samples
   BATCH_SIZE = num_mcmc_samples;
end
MH_X_PROP  = sqrt(0.6);  % Standard deviation of latent variable proposal
P_NOISE    = sqrt(0.6);  % Standard deviation of pseudo inputs proposal

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PARSE GRAPH information.
%
% x_ychildren: observed children of each latent. It contains the position
%              of each children wrt to the observed variables only, not 
%              the whole graph.
% x_xchildren: latent children of each latent. Again, numbers represent
%              position with respect the latent variables only.
% y_xparents:  latent parents of each observed variable (same indexing
%              comment applies).

num_nodes = size(graph, 1);
latents = find(n_types == 0);  observed = find(n_types == 1);
num_latents = length(latents); num_observed = length(observed);
children = cell(num_nodes, 1); parents = cell(num_nodes, 1);
for v = 1:num_nodes
    children{v} = find(graph(:, v) == 1)';
    parents{v} = find(graph(v, :) == 1);
end

x_ychildren = cell(num_latents, 1); 
x_xchildren = cell(num_latents, 1); x_xparents = cell(num_latents, 1);
exo_latent = zeros(num_latents, 1);
for x = 1:num_latents
    x_pos = latents(x);
    obs_children    = intersect(observed, children{x_pos});
    latent_children = intersect(latents, children{x_pos});
    latent_parents  = intersect(latents, parents{x_pos});
    x_ychildren{x} = zeros(1, length(obs_children));
    for v = 1:length(obs_children)
        x_ychildren{x}(v) = find(observed == obs_children(v), 1);
    end
    x_xchildren{x} = zeros(1, length(latent_children));
    for v = 1:length(latent_children)
        x_xchildren{x}(v) = find(latents == latent_children(v), 1);
    end
    x_xparents{x} = zeros(1, length(latent_parents));
    for v = 1:length(latent_parents)
        x_xparents{x}(v) = find(latents == latent_parents(v), 1);
    end
    exo_latent(x) = isempty(latent_parents);
end
exo_latents = find(exo_latent == 1)'; endo_latents = find(exo_latent == 0)';
y_xparents = cell(1, num_observed);
for y = 1:num_observed
    y_pos = observed(y);
    y_xparents{y} = zeros(1, length(parents{y_pos}));
    for v = 1:length(parents{y_pos})
        y_xparents{y}(v) = find(latents == parents{y_pos}(v), 1);
    end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ALLOCATE space. Everything is being done in RAM in batches, so this can 
% quickly exhauste the memory. Set BATCH_SIZE accordingly.

N  = size(data, 1);
sampled_x    = zeros(BATCH_SIZE + 1, N, num_latents);
sampled_f    = zeros(BATCH_SIZE + 1, N, num_latents);
sampled_fp   = zeros(BATCH_SIZE + 1, M, num_latents);
sampled_lin  = cell(num_observed, 1);
for v = 1:num_observed
    y = observed(v);
    sampled_lin{v} = zeros(BATCH_SIZE + 1, length(parents{y}) + 1);
end
sampled_xerr_var = zeros(BATCH_SIZE + 1, num_latents);
sampled_xmix     = cell(BATCH_SIZE + 1, length(exo_latents));
sampled_x_z      = zeros(BATCH_SIZE + 1, N, length(exo_latents));
sampled_yerr_par = zeros(BATCH_SIZE + 1, num_observed);
sampled_xp       = cell(BATCH_SIZE + 1, num_latents);

for x = exo_latents
    exo_pos = find(exo_latents == x, 1);
    num_k = length(hyper_x_mix{x, 1});
    for xi = 1:(BATCH_SIZE + 1)
        sampled_xmix{xi, exo_pos} = zeros(num_k, 3);
    end
end

% Cache all kernel matrices and decompositions
K_Mchol = cell(num_latents, 1);
K_NM    = cell(num_latents, 1); 
K_N     = cell(num_latents, 1);
alpha   = cell(num_latents, 1);
f_means = cell(num_latents, 1);

logtheta = cell(num_latents, 1);

pseudo_rho = zeros(1, BATCH_SIZE + 1);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% INITIAL SAMPLING. First, initialize error variances.
% After that, order latents in a partial order where a node 
% cannot come before any of its descendants. Then sample according to
% order.


if use_starting
   sampled_xerr_var(1, :) = starting_xvars;
   sampled_yerr_par(1, :) = starting_yvars;
   for d = 1:N
      sampled_x(1, d, :) = starting_latents(d, :);
   end
   for y = 1:num_observed
       sampled_lin{y}(1, :) = starting_loadings{y};
   end    
   sampled_kernel_hyper = cell(num_latents, 1);
   for x = 1:num_latents
       if ~isempty(x_xparents{x})
          sampled_kernel_hyper{x} = zeros(BATCH_SIZE + 1, ...
                           kernel_nparam(covfunc, length(x_xparents{x})));
          sampled_kernel_hyper{x}(1, :) = starting_kernelhyper{x}';
       end
   end
   for x = exo_latents
       exo_pos = find(exo_latents == x, 1);
       sampled_xmix{1, exo_pos} = starting_xmix{x};
       sampled_x_z(1, :, exo_pos) = starting_exo_z(x, :);
   end
   latent_order = intersect(get_node_order(graph), latents);
   pseudo_c = zeros(M, num_latents);
   for x_ord = 1:num_latents
       x = latent_order(x_ord);
       if isempty(x_xparents{x})
          continue
       end
       % Sample latent functions and pseudo latent functions        
       pseudo_c(:, x) = randsample(1:N, M);
       if length(x_xparents{x}) == 1
           x_p_data = sampled_x(1, :, x_xparents{x})';
           sampled_xp{1, x} = sampled_x(1, pseudo_c(:, x), x_xparents{x})' + ...
                                    P_NOISE * randn(M, length(x_xparents{x}));
       else
           x_p_data = squeeze(sampled_x(1, :, x_xparents{x}));
           sampled_xp{1, x} = squeeze(sampled_x(1, pseudo_c(:, x), x_xparents{x})) + ...
                                    P_NOISE * randn(M, length(x_xparents{x}));
       end
       logtheta{x} = kernel_gpml_map(covfunc, sampled_kernel_hyper{x}(1, :), 1);
       K_M         = feval(covfunc{:}, logtheta{x}, sampled_xp{1, x});   
       K_Mchol{x}  = chol(K_M);
       sampled_fp(1, :, x) = randnorm(1, zeros(M, 1), K_Mchol{x});
       [K_N{x} K_NM{x}] = feval(covfunc{:}, logtheta{x}, ...
                                sampled_xp{1, x}, x_p_data);
       alpha{x}   = K_Mchol{x}' \ K_NM{x}; % Here, K_NM is still a M x N matrix
       f_means{x} = alpha{x}' * (K_Mchol{x}' \ sampled_fp(1, :, x)');
       K_NM{x}    = K_NM{x}';              % Now a N x M, as it should be
       
       K_N{x} = K_N{x} - sum(alpha{x} .* alpha{x})';
       % Sample latent functions
       var_f = 1 ./ (1 ./ K_N{x} + 1 / sampled_xerr_var(1, x));
       mu_f  = var_f .* (f_means{x} ./ K_N{x} + ...
                         sampled_x(1, :, x)' / sampled_xerr_var(1, x));
       sampled_f(1, :, x) = mu_f + sqrt(var_f) .* randn(N, 1);        
   end
else
   sampled_kernel_hyper = cell(num_latents, 1);
   for x = 1:num_latents
       if ~isempty(x_xparents{x})
          sampled_kernel_hyper{x} = ones(BATCH_SIZE + 1, ...
                                       kernel_nparam(covfunc, length(x_xparents{x})));
          sampled_kernel_hyper{x}(:, fixed_hyper_kernel == 1) = ...
                                       repmat(hyper_kernel_values(fixed_hyper_kernel == 1)', BATCH_SIZE + 1, 1);
       end
   end
   sampled_xerr_var(1, :) = xerr';
   sampled_yerr_par(1, :) = yerr';
   latent_order = intersect(get_node_order(graph), latents);
   pseudo_c = zeros(M, num_latents);
   for x_ord = 1:num_latents
       x = latent_order(x_ord);
       if isempty(x_xparents{x})
          sampled_x(1, :, x) = randnorm(N, 1) * ...
                               sqrt(sampled_xerr_var(1, x));
          continue
       end
       % Sample latent functions and pseudo latent functions        
       pseudo_c(:, x) = randsample(1:N, M);
       if length(x_xparents{x}) == 1
           x_p_data = sampled_x(1, :, x_xparents{x})';
           sampled_xp{1, x} = sampled_x(1, pseudo_c(:, x), x_xparents{x})' + ...
                                    P_NOISE * randn(M, length(x_xparents{x}));
       else
           x_p_data = squeeze(sampled_x(1, :, x_xparents{x}));
           sampled_xp{1, x} = squeeze(sampled_x(1, pseudo_c(:, x), x_xparents{x})) + ...
                                    P_NOISE * randn(M, length(x_xparents{x}));
       end
       logtheta{x} = kernel_gpml_map(covfunc, sampled_kernel_hyper{x}(1, :), 1);
       K_M         = feval(covfunc{:}, logtheta{x}, sampled_xp{1, x});  
       K_Mchol{x}  = chol(K_M);
       sampled_fp(1, :, x) = randnorm(1, zeros(M, 1), K_Mchol{x});
       [K_N{x} K_NM{x}] = feval(covfunc{:}, logtheta{x}, ...
                                sampled_xp{1, x}, x_p_data);
       alpha{x}  = K_Mchol{x}' \ K_NM{x};
       f_means{x} = alpha{x}' * (K_Mchol{x}' \ sampled_fp(1, :, x)');
       K_NM{x}   = K_NM{x}';                            
       K_N{x}    = K_N{x} - sum(alpha{x} .* alpha{x})';
       % Sample latent functions
       var_f = 1 ./ (1 ./ K_N{x} + 1 / sampled_xerr_var(1, x));
       mu_f  = var_f .* (f_means{x} ./ K_N{x} + ...
                         sampled_x(1, :, x)' / sampled_xerr_var(1, x));
       sampled_f(1, :, x) = mu_f + sqrt(var_f) .* randn(N, 1);        
       % Sample latent data points
       sampled_x(1, :, x) = sampled_f(1, :, x) + randnorm(N, 1) * ...
                            sqrt(sampled_xerr_var(1, x));
   end
   for y = 1:num_observed % And now, sample linear coefficients
       n_par = length(y_xparents{y});
       cov_y = eye(n_par + 1);
       sampled_lin{y}(1, :) = randnorm(1, zeros(n_par + 1, 1), [], cov_y)';
   end
   for x = exo_latents
       exo_pos = find(exo_latents == x, 1);
       num_k = length(hyper_x_mix{x, 1});
       sampled_x_z(1, :, exo_pos) = sample(ones(num_k, 1) / num_k, N);
       for k = 1:num_k
           sampled_xmix{1, exo_pos}(k, 1) = 1 / num_k; % Probability 
           sampled_xmix{1, exo_pos}(k, 2) = randn();   % Mean
           sampled_xmix{1, exo_pos}(k, 3) = rand();    % Variance
       end
   end
end

for y = 1:num_observed
    n_par = length(y_xparents{y});
    sampled_lin{y}(1, fixed_loadings{y} == 1) = 1;
    if fixed_loadings{y}(n_par + 1)
       sampled_lin{y}(1, n_par + 1) = 0;
    end
end

% Initialize missing values
num_missing = size(missing_values, 1);
count_vars = N * ones(num_observed, 1);
mars_matrix = zeros(size(data));

for t = 1:num_missing
    r = missing_values(t, 1); c = missing_values(t, 2);
    data(r, c) = 0;
    count_vars(c) = count_vars(c) - 1;
end
data_means = sum(data) ./ count_vars';
if num_missing > 0
   sampled_missing = zeros(BATCH_SIZE + 1, num_missing);
else
   sampled_missing = [];
end
for t = 1:num_missing
    r = missing_values(t, 1); c = missing_values(t, 2);
    data(r, c) = data_means(c) + 0.01 * randn();
    sampled_missing(1, t) = data(r, c);
    mars_matrix(r, c) = 1;
end

% Initialize information about pseudo-inputs prior
PSEUDO_PRIOR = 2; % 0: independent Gaussians 
switch PSEUDO_PRIOR
    case 1  % 1: anticorrelation prior     
        pseudo_rho(1) = -0.5;
        min_rho       = -0.5;
        template_rho  = zeros(M, M); for ti = 1:(M - 1), template_rho(ti, ti + 1) = 1; end; for ti = 2:M, template_rho(ti, ti - 1) = 1; end;
        rho_matrix    = eye(M) + pseudo_rho(1) * template_rho;
        pseudo_mean   = zeros(M, M - 1);
        pseudo_var    = zeros(M, 1);
        for pi = 1:M
            other_m          = [1:(pi - 1) (pi + 1):M];
            rho_matrix_pi    = rho_matrix(other_m, other_m);
            coeff_pi         = zeros(1, M - 1);
            if pi > 1, coeff_pi(pi - 1) = pseudo_rho(1); end
            if pi < M, coeff_pi(pi) = pseudo_rho(1); end
            pseudo_mean(pi, :)  = coeff_pi * inv(rho_matrix_pi);
            pseudo_var(pi)      = P_VARIANCE - pseudo_mean(pi, :) * (coeff_pi' * P_VARIANCE);
        end
    case 2 % 2: determinant prior
        MAX_XP  = 5;
        XP_PROP = 1;
        covfunc_xp_prior  = {'covSum', {'covSEiso','covNoise'}};
        logtheta_xp_prior = kernel_gpml_map(covfunc_xp_prior, [0.1; 1; 1.e-4], 1);
        K_xpchol          = cell(num_latents, 1);
        for x = 1:num_latents
            if ~isempty(parents{x})
               in_data = sampled_xp{1, x};
               K_xpchol{x} = chol(feval(covfunc_xp_prior{:}, logtheta_xp_prior, in_data));
            end
        end
end

disp(sampled_yerr_par(1, :));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SETUP BATCH SIZE INFORMATION

num_batches = ceil(num_mcmc_samples / BATCH_SIZE);
num_batch_samples = BATCH_SIZE * ones(num_batches, 1);
num_batch_samples(num_batches) = num_mcmc_samples - BATCH_SIZE * (num_batches - 1);

file_x       = fopen('sample_x.txt','w');
file_f       = fopen('sample_f.txt','w');
file_pseudox = fopen('sample_pseudox.txt','w');
file_pseudof = fopen('sample_pseudof.txt','w');
file_exo     = fopen('sample_exo.txt', 'w');
file_xvar    = fopen('sample_xvar.txt', 'w');
file_y       = fopen('sample_y.txt','w');
file_z       = fopen('sample_z.txt','w');
file_hgp     = fopen('sample_hgp.txt','w');
file_lambda  = fopen('sample_lambda.txt','w');
file_yvar    = fopen('sample_yvar.txt','w');

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PERFORM SAMPLING

total_iter = 0;

for bt = 1:num_batches
for iter = 2:(num_batch_samples(bt) + 1)

    total_iter = total_iter + 1; disp(sprintf('Iteration %d', total_iter));

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Initialization of iteration 
    
    sampled_x(iter, :, :)     = sampled_x(iter - 1, :, :);
    sampled_xerr_var(iter, :) = sampled_xerr_var(iter - 1, :);
    for x = 1:num_latents
        if ~isempty(x_xparents{x})
           sampled_kernel_hyper{x}(iter, :) = sampled_kernel_hyper{x}(iter - 1, :);
           sampled_f(iter, :, x)            = sampled_f(iter - 1, :, x);
           sampled_fp(iter, :, x)           = sampled_fp(iter - 1, :, x);
           sampled_xp{iter, x}              = sampled_xp{iter - 1, x};
        end        
    end
    sampled_yerr_par(iter, :) = sampled_yerr_par(iter - 1, :);
    for y = 1:num_observed        
        sampled_lin{y}(iter, :) = sampled_lin{y}(iter - 1, :);
    end

    pseudo_rho(iter) = pseudo_rho(iter - 1);
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Sample latent functions (``pseudo'' and ``actual'') given latent 
    % variables and data

    for x = 1:num_latents
        if ~isempty(parents{x})
            % Sample pseudo functions
            inv_L  = inv(K_Mchol{x});
            invK_M = inv_L * inv_L';
            coeffs = K_NM{x} * invK_M;
            aux_x  = coeffs' * diag(1./(K_N{x} + sampled_xerr_var(iter, x)));
            Sigma_fp = inv(invK_M + aux_x * coeffs);
            Sigma_fp = 0.5 * (Sigma_fp + Sigma_fp'); % For numerical reasons
            mean_fp  = Sigma_fp * aux_x * sampled_x(iter, :, x)';
            sampled_fp(iter, :, x) = randnorm(1, mean_fp, [], Sigma_fp);
            % Update f_means
            f_means{x} = alpha{x}' * (K_Mchol{x}' \ sampled_fp(iter, :, x)');
            % Sample ``actual'' functions given pseudo-functions f
            var_f = 1 ./ (1 ./ K_N{x} + 1 / sampled_xerr_var(iter, x));
            mu_f  = var_f .* (f_means{x} ./ K_N{x} + ...
                              sampled_x(iter, :, x)' / sampled_xerr_var(iter, x));
            sampled_f(iter, :, x) = mu_f + sqrt(var_f) .* randn(N, 1);
        end
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Sample kernel hyperparameters

    for x = 1:num_latents
        if ~isempty(parents{x})
           if length(x_xparents{x}) == 1
              x_data = sampled_x(iter, :, x_xparents{x})';
           else
              x_data = squeeze(sampled_x(iter, :, x_xparents{x}));
           end
           pseudo_data = sampled_xp{iter, x};
           np = kernel_nparam(covfunc, length(x_xparents{x}));           
           k_format0 = ' %.4f'; k_format = k_format0; for kf = 2:np, k_format = strcat(k_format, k_format0); end

           % Propose hyperparameter values
           curr_par = sampled_kernel_hyper{x}(iter, :);
           new_par = curr_par;
           logfwd = 0; logbkw = 0; log_currprior = 0; log_newprior = 0;
           for p = 1:np
               if fixed_hyper_kernel(p)
                  continue
               end
               [new_par_p logfwd_p logbkw_p] = kernel_prop(covfunc, p, curr_par(p));
               new_par(p) = new_par_p;
               logfwd = logfwd + logfwd_p;
               logbkw = logbkw + logbkw_p;
               log_newprior  = log_newprior  + kernel_logprior(covfunc, p, new_par(p));
               log_currprior = log_currprior + kernel_logprior(covfunc, p, curr_par(p));
           end

           old_logtheta = logtheta{x};
           logtheta{x}= kernel_gpml_map(covfunc, new_par, 1);
           newK_M     = feval(covfunc{:}, logtheta{x}, pseudo_data);  
           newK_Mchol = chol(newK_M);
           [newK_N newK_NM] = feval(covfunc{:}, logtheta{x}, pseudo_data, x_data);
           new_alpha  = newK_Mchol' \ newK_NM;
           new_fmean  = new_alpha' * (newK_Mchol' \ sampled_fp(iter, :, x)');
           newK_NM    = newK_NM';
           newK_N     = newK_N - sum(new_alpha .* new_alpha)';
           
           fmean      = f_means{x};
           f_data     = sampled_f(iter, :, x)';
           fp_data    = sampled_fp(iter, :, x)';
           new_f_norm = newK_Mchol' \ fp_data; new_f_norm = sum(new_f_norm .* new_f_norm);
           old_f_norm = K_Mchol{x}' \ fp_data; old_f_norm = sum(old_f_norm .* old_f_norm);
           new_logfactor = -0.5 * (2 * sum(log(diag(newK_Mchol))) + new_f_norm + ...
                               sum(log(newK_N)) + ...
                               sum((f_data - fmean).^2 ./ newK_N)) + ... 
                            log_newprior;
           old_logfactor = -0.5 * (2 * sum(log(diag(K_Mchol{x}))) + old_f_norm + ...
                               sum(log(K_N{x})) + ...
                               sum((f_data - new_fmean).^2 ./ K_N{x})) + ... 
                            log_currprior;
           
           ratio = exp(new_logfactor + logbkw - old_logfactor - logfwd);

           if rand() < ratio % Accept it 
              K_N{x}     = newK_N;
              K_NM{x}    = newK_NM;
              f_means{x} = new_fmean;
              alpha{x}   = new_alpha;
              K_Mchol{x} = newK_Mchol;
              sampled_kernel_hyper{x}(iter, :) = new_par;              
           else
              logtheta{x} = old_logtheta;
           end                       
           %disp(sampled_kernel_hyper{x}(iter, :));
           disp(sprintf(strcat(k_format, ' - Mean K_N{%d}: %f'), sampled_kernel_hyper{x}(iter, :), x, mean(K_N{x})));
        end
    end

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Sample latent variable conditional variance hyperparameters
    % (for endogenous latent variables)
    
    for x = endo_latents
        if fixed_xerr(x)
           continue
        end
        ss_x = sum((sampled_x(iter, :, x) - sampled_f(iter, :, x)) .^ 2);
        ga = hyper_x_var(1) + N / 2;
        gb = hyper_x_var(2) + ss_x / 2;
        sampled_xerr_var(iter, x) = gb / randgamma(ga);
    end

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Sample latent variable marginal (mixture of Gaussians) parameters
    % (for exogenous latent variables)
    
    for x = exo_latents        
        exo_pos = find(exo_latents == x, 1);
        num_k = length(hyper_x_mix{x, 1});
        % First, sample means and variances for each mixture component        
        for k = 1:num_k
            k_rows = find(sampled_x_z(iter - 1, :, exo_pos) == k);
            s_x = sum(sampled_x(iter - 1, k_rows, x));            
            % Sample mean give variance
            v_mean = 1 / (1 / hyper_x_mix{x, 2} + ...
                     length(k_rows) / sampled_xmix{iter - 1, exo_pos}(k, 3));
            m_mean = v_mean * s_x / sampled_xmix{iter - 1, exo_pos}(k, 3);
            new_mean = randn() * sqrt(v_mean) + m_mean;
            sampled_xmix{iter, exo_pos}(k, 2) = new_mean;
            % Sample variance given mean
            ss_x = sum((sampled_x(iter - 1, k_rows, x) - new_mean).^2);
            ga = hyper_x_mix{x, 3}(1) + length(k_rows) / 2;
            gb = hyper_x_mix{x, 3}(2) + ss_x / 2;
            sampled_xmix{iter, exo_pos}(k, 3) = gb / randgamma(ga);
        end
        % Now sample mixture component indicators given x and means
        for d = 1:N
            weights = zeros(num_k, 1);
            for k = 1:num_k
                m = sampled_xmix{iter, exo_pos}(k, 2);
                v = sampled_xmix{iter, exo_pos}(k, 3);
                xd = sampled_x(iter - 1, d, x);
                weights(k) = -0.5 * log(v) - 0.5 * (xd - m)^2 / v + ...
                             log(sampled_xmix{iter - 1, exo_pos}(k, 1));
            end
            weights = weights - max(weights); weights = exp(weights);
            sampled_x_z(iter, d, exo_pos) = sample(weights/sum(weights));
        end
        % Finally, sample mixture distribution
        alpha_m = hyper_x_mix{x, 1};
        for k = 1:num_k
            alpha_m(k) = alpha_m(k) + sum(sampled_x_z(iter, :, exo_pos) == k);
        end
        new_prob = randdirichlet(alpha_m);
        for k = 1:num_k
            sampled_xmix{iter, exo_pos}(k, 1) = new_prob(k);
        end        
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Sample measurement error hyperparameters
    
    for y = 1:num_observed
        if fixed_yerr(y)
           continue
        end
        recN  = mars_matrix(:, y) == 0;
        if length(y_xparents{y}) > 1
           x_data = [squeeze(sampled_x(iter, recN, y_xparents{y})) repmat(1, sum(recN), 1)];
        else
           x_data = [sampled_x(iter, recN, y_xparents{y})' repmat(1, sum(recN), 1)]; 
        end        
        ss_y = sum((data(recN, y) - x_data * sampled_lin{y}(iter, :)') .^ 2);
        ga = hyper_y_par(1) + sum(recN) / 2;
        gb = hyper_y_par(2) + ss_y / 2;
        sampled_yerr_par(iter, y) = gb / randgamma(ga);
    end
    disp(sampled_yerr_par(iter, :));
        
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Sample linear coefficients given latent variables and data

    for y = 1:num_observed        
        recN  = mars_matrix(:, y) == 0;
        n_par = length(y_xparents{y});
        if length(y_xparents{y}) > 1
           x_data = [squeeze(sampled_x(iter, recN, y_xparents{y})) repmat(1, sum(recN), 1)];
        else
           x_data = [sampled_x(iter, recN, y_xparents{y})' repmat(1, sum(recN), 1)];
        end
        cov_y = x_data' * x_data / sampled_yerr_par(iter, y);
        for yi = 1:(n_par + 1)
            cov_y(yi, yi) = cov_y(yi, yi) + 1 / lin_var;
        end
        cov_y = inv(cov_y);
        mean_y = cov_y * sum(repmat(data(recN, y), 1, n_par + 1) .* x_data)' / ...
                         sampled_yerr_par(iter, y); 
        if sum(fixed_loadings{y} == 0)
           sampled_lin{y}(iter, :) = randnorm(1, mean_y, [], cov_y)';        
        else
           fr = fixed_loadings{y} == 0; % Free coefficients
           fx = fixed_loadings{y} == 1; % Fixed coefficients
           if ~isempty(fixed_loading_values)
              sampled_lin{y}(iter, fx) = fixed_loading_values{y}';
           else
              sampled_lin{y}(iter, fx) = 1;            
              if fixed_loadings{y}(n_par + 1)
                 sampled_lin{y}(iter, n_par + 1) = 0;
              end
           end            
           reg_coeff = cov_y(fr, fx) * inv(cov_y(fx, fx)); 
           mean_f = reg_coeff * (sampled_lin{y}(iter, fx)' - mean_y(fx));
           cov_f  = cov_y(fr, fr) - reg_coeff * cov_y(fx, fr);
           sampled_lin{y}(iter, fr) = randnorm(1, mean_f, [], cov_f)';
        end
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Sample latent variables
    
    sample_latents();

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Sample missing values
    
    if num_missing > 0
       sds = sqrt(sampled_yerr_par(iter, :));
       rand_pool = randn(num_missing, 1);
       for t = 1:num_missing
           d = missing_values(t, 1);
           y = missing_values(t, 2);        
           m_t = sampled_lin{y}(iter, :) * [sampled_x(iter, d, y_xparents{y}); 1];
           data(d, y) = m_t + sds(y) * rand_pool(t);
           sampled_missing(iter, t) = m_t; % Report mean only
       end
    end

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Sample pseudo-data
    
    sample_pseudo_x();
    if verbose > 4
       print_coeff();
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Sample pseudo-data correlation
    if PSEUDO_PRIOR == 1
       update_pseudoprior()
    end
    
end % batch

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% NEXT BATCH preparation

last_s = num_batch_samples(bt) + 1;
print_to_file(last_s); % Print whole batch to file
sampled_x(1, :, :) = sampled_x(last_s, :, :);
sampled_f(1, :, :) = sampled_f(last_s, :, :);
for v = 1:num_observed
    sampled_lin{v}(1, :) = sampled_lin{v}(last_s, :);
end
sampled_xerr_var(1, :) = sampled_xerr_var(last_s, :);
for x = exo_latents
    exo_pos = find(exo_latents == x, 1);
    sampled_xmix{1, exo_pos} = sampled_xmix{last_s, exo_pos};
end
sampled_x_z(1, :, :)   = sampled_x_z(last_s, :, :);
sampled_yerr_par(1, :) = sampled_yerr_par(last_s, :);

sampled_fp(1, :, :) = sampled_fp(last_s, :, :);
for x = 1:num_latents
    if ~isempty(parents{x})
       sampled_kernel_hyper{x}(1, :) = sampled_kernel_hyper{x}(last_s, :);
       sampled_xp{1, x} = sampled_xp{last_s, x};
    end
end
if num_missing > 0
   sampled_missing(1, :) = sampled_missing(last_s, :);
end

pseudo_rho(1) = pseudo_rho(last_s);

end % all batches


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CLEAN-UP

fclose(file_x); fclose(file_f); fclose(file_exo); fclose(file_xvar);
fclose(file_y); fclose(file_z); fclose(file_hgp);
fclose(file_lambda); fclose(file_yvar); 
fclose(file_pseudof); fclose(file_pseudox); 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Local SAMPLING functions

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Hard part: change the pseudo points.

function sample_pseudo_x()
    if verbose > 0; disp('Sampling pseudo latents... '); end
    
    num_accepted = zeros(num_latents, 1);
    for xx = 1:num_latents        
        if isempty(parents{xx})
           continue
        end
        
        perm_M = randperm(M);
        for xm = perm_M %#ok<FXUP>
            if verbose > 1, disp(sprintf('Sampled pseudo %d', xm)); end
            other_m   = [1:(xm - 1) (xm + 1):M];            
            cond_data = sampled_xp{iter, xx}(other_m, :);
            
			%%% Do fast update for prior factor, if necessary
            if PSEUDO_PRIOR == 2
               L_sub_xp           = subchol(K_xpchol{xx}, xm);            
               pred_data = sampled_xp{iter, xx}(xm, :);
               [kk_starp k_starp] = feval(covfunc_xp_prior{:}, logtheta_xp_prior, cond_data, pred_data);
               vv_xp              = L_sub_xp' \ k_starp;
               logdet_factor      = log(kk_starp - sum(vv_xp .* vv_xp));
            else
               logdet_factor = [];
            end
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            %%%%% Collect factors for current indicator
            old_sampled_xp = sampled_xp{iter, xx}(xm, :);
            backup_KMchol = K_Mchol{xx};
            backup_KNM    = K_NM{xx}(:, xm);
            backup_KN     = K_N{xx};
            backup_fmeans = f_means{xx};
            backup_alpha  = alpha{xx};
            logfactor1    = xp_logfactor(xx, xm, other_m, logdet_factor);

            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            %%%%% Propose new pseudoinput
            if PSEUDO_PRIOR ~= 2
               sampled_xp{iter, xx}(xm, :) = old_sampled_xp + ...
                                             P_NOISE * randn(1, length(x_xparents{xx}));
               logfwd = 0; logbwd = 0;
            else
               [sampled_xp{iter, xx}(xm, :) logfwd logbwd] = ...
                                             propose_pseudoinput(old_sampled_xp);
            end
            pred_data = sampled_xp{iter, xx}(xm, :);
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            %%%%% Propose and collect factors for pseudo function f_{xm}
            
            [L_sub alpha_sub] = subchol(K_Mchol{xx}, xm, alpha{xx});
            
            %%% Get predictive mean                       
            [kk_star k_star] = feval(covfunc{:}, logtheta{xx}, cond_data, pred_data);
            alpha_p   = solve_chol(L_sub, sampled_fp(iter, other_m, xx)');
            pred_mean = k_star' * alpha_p;
            %%% Get predictive variance
            vv = L_sub' \ k_star;
            pred_var = kk_star - vv' * vv;
            %%% Do actual resample    
            old_sampled_fp = sampled_fp(iter, xm, xx);
            sampled_fp(iter, xm, xx) = randn() * sqrt(pred_var) + pred_mean;
            
			%%% Do fast update for prior factor, if necessary            
            if PSEUDO_PRIOR == 2
               [kk_starp k_starp] = feval(covfunc_xp_prior{:}, logtheta_xp_prior, cond_data, pred_data);
               vv_xp              = L_sub_xp' \ k_starp;
               logdet_factor      = log(kk_starp - sum(vv_xp .* vv_xp));
            end
            
			%%% Get updated kernels, part I
            if length(x_xparents{xx}) == 1
               x_data = sampled_x(iter, :, x_xparents{xx})';
            else
               x_data = squeeze(sampled_x(iter, :, x_xparents{xx}));
            end
            [new_k_dd new_k_dp] = feval(covfunc{:}, logtheta{xx}, ...
                                        sampled_xp{iter, xx}(xm, :), x_data);

            %%% Re-calculate Cholesky factors, etc.
            K_Mchol{xx}(1:(M - 1), 1:(M - 1)) = L_sub;
            alpha{xx}(1:(M - 1), :) = alpha_sub;
            cholroll(K_Mchol{xx}, [k_star' kk_star], alpha{xx}, new_k_dp); % Update Cholesky/alpha to include new row/column
            if xm < M
               cholupexch(K_Mchol{xx}, xm, M, 1, alpha{xx});               % Reorganize Cholesky to reflect original order
            end

            %%% Get updated kernels, part II
            K_NM{xx}(:, xm)  = new_k_dp';
            f_means{xx}      = alpha{xx}' * (K_Mchol{xx}' \ sampled_fp(iter, :, xx)');
            K_N{xx}          = new_k_dd - sum(alpha{xx} .* alpha{xx})';
            
            %%% Get new factor            
            logfactor2 = xp_logfactor(xx, xm, other_m, logdet_factor);
        
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % DEBUG CODE TO COMPARE THE FAST UPDATES TO THE FROM-SCRATCH
            % ONES
            %dK_M     = feval(covfunc{:}, logtheta{xx}, sampled_xp{iter, xx});   
            %dK_Mchol = chol(dK_M);
            %[dK_N dK_NM] = feval(covfunc{:}, logtheta{xx}, ...
            %                     sampled_xp{iter, xx}, x_data);
            %dalpha   = dK_Mchol' \ dK_NM;
            %df_means = dalpha' * (dK_Mchol' \ sampled_fp(iter, :, xx)');
            %dK_NM    = dK_NM'; 
            %dK_N     = dK_N - sum(dalpha .* dalpha)';
            %error_chol = max(max(abs(dK_Mchol - K_Mchol{xx})));
            %error_NM   = max(max(abs(dK_NM - K_NM{xx})));
            %error_N    = max(abs(dK_N - K_N{xx}));
            %error_fm   = max(max(abs(df_means - f_means{xx})));
            %error_al   = max(max(abs(dalpha - alpha{xx})));
            %monitor    = max([error_chol error_NM error_N error_fm error_al]);
            %disp(monitor);
            %if monitor > 1.e-4
            %   error('Unacceptable numerical error!');
            %end
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            %%% Evaluate proposal
            ratio = exp(logfactor2 + logbwd - logfactor1 - logfwd);
            if rand() > ratio % Reject it
               sampled_xp{iter, xx}(xm, :) = old_sampled_xp;
               sampled_fp(iter, xm, xx)    = old_sampled_fp;
               K_Mchol{xx}     = backup_KMchol;
               K_NM{xx}(:, xm) = backup_KNM;
               K_N{xx}         = backup_KN;
               f_means{xx}     = backup_fmeans;
               alpha{xx}       = backup_alpha;
            else
               num_accepted(xx) = num_accepted(xx) + 1;
               %%% Re-calculate Cholesky factors for pseudo-input prior, if necessary
               if PSEUDO_PRIOR == 2
                  K_xpchol{xx}(1:(M - 1), 1:(M - 1)) = L_sub_xp;
                  cholroll(K_xpchol{xx}, [k_starp' kk_starp]); % Update Cholesky/alpha to include new row/column
                  if xm < M
                     cholupexch(K_xpchol{xx}, xm, M, 1);       % Reorganize Cholesky to reflect original order
                  end
               end
            end
        end

    end
    disp(num_accepted');
    if verbose > 0, disp('Sampling pseudo latents... DONE'); end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Finally, sample latent variables given latent functions, 
% linear coefficients and data.

function sample_latents()
    if verbose > 0, disp('Sampling latents... '); end

    % Get all proposals
    old_xv = squeeze(sampled_x(iter - 1, :, :));
    if num_latents == 1
       old_xv = old_xv';
    end
    new_xv = randn(N, num_latents) * MH_X_PROP + old_xv;
    
    % All of such datapoints are independent of each other given the
    % pseudo-inputs
    for xx = 1:num_latents
        log_factors1 = x_logfactor(xx);
        sampled_x(iter, :, xx) = new_xv(:, xx);
        update_crosskernels_fast(xx, 1:N);
        old_xvc = sampled_f(iter, :, x_xchildren{xx});
        sampled_f(iter, :, x_xchildren{xx}) = propose_latent_fchildren(xx);
        log_factors2 = x_logfactor(xx);
        ratios = exp(log_factors2 - log_factors1);
        rn = rand(N, 1);
        rejects = rn > ratios == 1;
        sampled_x(iter, rejects, xx) = old_xv(rejects, xx);
        if sum(rejects) > 0
           update_crosskernels_fast(xx, rejects);
           sampled_f(iter, rejects, x_xchildren{xx}) = old_xvc(1, rejects, :);
        end
        %disp(sprintf('num rejects: %d (out of %d)', length(rejects), length(ratios)));
    end
    
    if verbose > 0, disp('Sampling latents... DONE'); end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Local AUXILIARY functions

% This 'fast' variation knows that pseudo-inputs have not changed. Hence, 
% all of these points are independent conditioned on the pseudo-inputs, 
% making joint computation advantageous.
%
% Also, propose latent children functions.

function logfactors = x_logfactor(xx)
    % Factor due to latent parent function/marginal
    if ~exo_latent(xx)
       logfactors = (-0.5 * (sampled_x(iter, :, xx) - sampled_f(iter, :, xx)).^2 ./ ...
                             sampled_xerr_var(iter, xx))';
    else       
       exo_pos = find(exo_latents == xx, 1);
       z_x = sampled_x_z(iter, :, exo_pos);
       num_k = length(hyper_x_mix{xx, 1});
       mms = zeros(N, 1); vrs = zeros(N, 1);
       for zz = 1:num_k
           z_idx = find(z_x == zz);
           mms(z_idx) = sampled_xmix{iter, exo_pos}(zz, 2);
           vrs(z_idx) = sampled_xmix{iter, exo_pos}(zz, 3);
       end
       logfactors = -0.5 * (sampled_x(iter, :, xx)' - mms).^2 ./ vrs;
    end
    % Factors due to observed (and recorded) children
    for y_c = x_ychildren{xx}
        recN  = (mars_matrix(:, y_c) == 0);
        if length(y_xparents{y_c}) > 1
           x_data = [squeeze(sampled_x(iter, recN, y_xparents{y_c})) repmat(1, sum(recN), 1)];
        else
          x_data = [sampled_x(iter, recN, y_xparents{y_c})' repmat(1, sum(recN), 1)];
        end
        b_y = sampled_lin{y_c}(iter, :)';
        logfactors(recN) = logfactors(recN) - ...
                     0.5 * (data(recN, y_c) - x_data * b_y).^2 ./ sampled_yerr_par(iter, y_c);
    end
    % Factors due to latent children functions 
    for x_c = x_xchildren{xx}
        % Not a pseudo-input, so only factor of the diagonal ``GP''        
        logfactors = logfactors - ...                     
                     0.5 * (sampled_x(iter, :, x_c) - sampled_f(iter, :, x_c))'.^2 ./ sampled_xerr_var(iter, x_c);
    end
end

% Propose values for children latent functions

function fs = propose_latent_fchildren(xx)
    fs = zeros(N, length(x_xchildren{xx}));
    % Factors due to latent children functions 
    i = 1;
    for x_c = x_xchildren{xx}
        fs(:, i) = randn(N, 1) .* sqrt(K_N{x_c}) + f_means{x_c};
        i        = i + 1;
    end
end

% This version knows points in 'D' are not a pseudo-input to any of its
% children. Update everything simultaneously. Cost is O(NM^2).

function update_crosskernels_fast(xx, D)
    for x_c = x_xchildren{xx}  
        if length(x_xparents{x_c}) == 1
           x_data = sampled_x(iter, D, x_xparents{x_c})';
        else            
           x_data = squeeze(sampled_x(iter, D, x_xparents{x_c}));
        end
        [new_k_dd new_k_dp] = feval(covfunc{:}, logtheta{x_c}, ...
                                    sampled_xp{iter, x_c}, x_data);
        alpha{x_c}(:, D) = K_Mchol{x_c}' \ new_k_dp;
        K_NM{x_c}(D, :)  = new_k_dp';        
        f_means{x_c}(D)  = alpha{x_c}(:, D)' * (K_Mchol{x_c}' \ sampled_fp(iter, :, x_c)');
        K_N{x_c}(D)      = new_k_dd - sum(alpha{x_c}(:, D) .* alpha{x_c}(:, D))';
    end
end

% Pseudopoint logfactor
function logfactor = xp_logfactor(xx, m, other_m, logdet_factor)
    % Factor due to the prior of the pseudo input    
    switch PSEUDO_PRIOR
        case 0 % WHEN PRIOR IS INDEPENDENT GAUSSIAN
           logfactor = - 0.5 * sum(sampled_xp{iter, xx}(m, :).^2) / P_VARIANCE;
        case 1 % WHEN PRIOR IS NEGATIVELY CORRELATED GAUSSIAN
           logfactor = -0.5 * sum((sampled_xp{iter, xx}(m, :) - pseudo_mean(m, :) * sampled_xp{iter, xx}(other_m, :)).^2) / pseudo_var(m);
        case 2 % WHEN PRIOR IS DETERMINANT GP
           logfactor = logdet_factor;
    end

    % Factors coming from the ``true'' latent functions and their ``diagonal GP''
    % All factors change, since "f_means" depends on all pseudoinputs.
    logfactor = logfactor - 0.5 * sum(log(K_N{xx})) - ...
                0.5 * sum((sampled_f(iter, :, xx)' - f_means{xx}).^2 ./ K_N{xx});
end

function [new_xp logfwd logbwd] = propose_pseudoinput(xp)
   s = length(xp);
   new_xp = zeros(s, 1);
   logfwd = 0;
   logbwd = 0;
   for ss = 1:s
       a = max(-MAX_XP, xp(ss) - XP_PROP);
       b = min(MAX_XP, xp(ss) + XP_PROP);
       new_xp(ss) = a + rand() * (b - a);
       logfwd = logfwd -log(b - a);
       a = max(-MAX_XP, new_xp(ss) - XP_PROP);
       b = min(MAX_XP, new_xp(ss) + XP_PROP);
       logbwd = logbwd -log(b - a);
   end
end

function update_pseudoprior()
   rho_a = min(0, pseudo_rho(iter) + 0.2);
   rho_b = max(min_rho, pseudo_rho(iter) - 0.2);
   fwd_prop = 1 / (rho_a - rho_b);
   rho_new = rho_b + rand() * (rho_a - rho_b); % Propose
   rho_ar = min(0, rho_new + 0.2);
   rho_br = max(min_rho, rho_new - 0.2);
   bwd_prop = 1 / (rho_ar - rho_br);
   % Collect factors
   logfactor_rho_old = 0; logfactor_rho_new = 0;
   rho_matrix     = eye(M) + template_rho * pseudo_rho(iter);
   invRho         = inv(rho_matrix);
   logdetrho_old  = logdet(rho_matrix);
   rho_matrix_new = eye(M) + template_rho * rho_new;
   invRhoNew      = inv(rho_matrix_new);
   logdetrho_new  = logdet(rho_matrix_new);
   for xx = 1:num_latents
       if ~isempty(parents{xx})
           logfactor_rho_old = logfactor_rho_old - 0.5 * length(parents{xx}) * logdetrho_old;
           for xp = 1:length(parents{xx})
               logfactor_rho_old = logfactor_rho_old - 0.5 * ...
                                   sampled_xp{iter, xx}(:, xp)' * invRho * sampled_xp{iter, xx}(:, xp) / P_VARIANCE;
           end
           logfactor_rho_new = logfactor_rho_new - 0.5 * length(parents{xx}) * logdetrho_new;
           for xp = 1:length(parents{xx})
               logfactor_rho_new = logfactor_rho_new - 0.5 * ...
                                   sampled_xp{iter, xx}(:, xp)' * invRhoNew * sampled_xp{iter, xx}(:, xp) / P_VARIANCE;
           end
       end
   end
   ratio = exp(logfactor_rho_new - logfactor_rho_old) * bwd_prop / fwd_prop;
   if rand() < ratio % Accept it 
      pseudo_rho(iter) = rho_new;
      for ppi = 1:M
          other_m          = [1:(ppi - 1) (ppi + 1):M];
          rho_matrix_pi    = rho_matrix_new(other_m, other_m);
          coeff_pi         = zeros(1, M - 1);
          if ppi > 1, coeff_pi(ppi - 1) = rho_new; end
          if ppi < M, coeff_pi(ppi)     = rho_new; end
          pseudo_mean(pi, :) = coeff_pi * inv(rho_matrix_pi);
          pseudo_var(pi)     = P_VARIANCE - pseudo_mean(ppi, :) * (coeff_pi' * P_VARIANCE);
      end
   end  
end

% Print coefficients (for debugging purposes)

function print_coeff()
    for yy = 1:num_observed
        disp(sampled_lin{yy}(iter, :));
    end
end

% Print all points in the batch to the respective files

function print_to_file(endpoint)
   disp('Printing to file...');
   % Dump latent variables
   x_format = ' %.3f';
   x_format_r = x_format;
   for w = 1:(num_latents - 1)
       x_format_r = strcat(x_format_r, x_format);
   end
   x_format_r = strcat(x_format_r, '\n');
   for s = 2:endpoint
       fprintf(file_x, x_format_r, sampled_x(s, :, :));    
   end
   % Dump imputed data
   if num_missing > 0
      y_format = ' %.3f';
      y_format_r = y_format;
      for w = 1:(num_missing - 1)
          y_format_r = strcat(y_format_r, y_format);
      end
      y_format_r = strcat(y_format_r, '\n');
      for s = 2:endpoint
          fprintf(file_y, y_format_r, sampled_missing(s, :));    
      end
   end
   % Dump exogenous mixture indicators
   z_format = ' %d';
   z_format_r = z_format;
   for w = 2:length(exo_latents)
       z_format_r = strcat(z_format_r, z_format);
   end
   z_format_r = strcat(z_format_r, '\n');
   for s = 2:endpoint
       fprintf(file_z, z_format_r, sampled_x_z(s, :, :));    
   end
   % Dump GP hyperparameters
   hgp_format = ' %.3f';
   for s = 2:endpoint
       for cl = 1:num_latents
           if isempty(parents{cl})
              continue
           end
           hgp_format_r = hgp_format;
           for w = 1:(length(sampled_kernel_hyper{cl}(s, :)) - 1)
               hgp_format_r = strcat(hgp_format_r, hgp_format);
           end
           fprintf(file_hgp, hgp_format_r, sampled_kernel_hyper{cl}(s, :));    
       end
       fprintf(file_hgp, '\n');    
   end   
   % Dump function values
   f_format_r = x_format_r;
   for s = 2:endpoint
       fprintf(file_f, f_format_r, sampled_f(s, :, :));    
   end
   % Dump pseudofunction values
   f_format_r = x_format_r;
   for s = 2:endpoint
       fprintf(file_pseudof, f_format_r, sampled_fp(s, :, :));    
   end
   % Dump latent variances
   xvar_format = ' %.3f';
   for s = 2:endpoint
       for xx = 1:num_latents
           fprintf(file_xvar, xvar_format, sampled_xerr_var(s, xx));    
       end
       fprintf(file_xvar, '\n');    
   end   
   % Dump pseudo points   
   xpvar_format = ' %.3f'; xpvar_format_r = xpvar_format;
   for gg = 2:M
       xpvar_format_r = strcat(xpvar_format_r, xpvar_format);
   end
   for s = 2:endpoint
       for xx = 1:num_latents
           if isempty(parents{xx})
              continue
           end
           for pp = 1:length(x_xparents{xx})
               fprintf(file_pseudox, xpvar_format_r, sampled_xp{s, xx}(:, pp));    
           end
           fprintf(file_pseudox, '\n');
       end       
   end   
   % Dump exogenous information
   exo_format = ' %.3f';
   for s = 2:endpoint
       for xx = 1:num_latents
           if ~isempty(parents{xx})
              continue
           end
           exo_pos = find(exo_latents == xx, 1);
           num_k = length(hyper_x_mix{xx, 1});
           for zz = 1:num_k
               fprintf(file_exo, exo_format, sampled_xmix{s, exo_pos}(zz, 1)); 
               fprintf(file_exo, exo_format, sampled_xmix{s, exo_pos}(zz, 2)); 
               fprintf(file_exo, exo_format, sampled_xmix{s, exo_pos}(zz, 3)); 
           end
       end
       fprintf(file_exo, '\n');    
   end   
   % Dump linear coefficient hyperparameters
   lambda_format = ' %.3f';
   for s = 2:endpoint
       for yy = 1:num_observed
           lambda_format_r = lambda_format;
           for w = 1:length(y_xparents{yy})
               lambda_format_r = strcat(lambda_format_r, lambda_format);
           end
           fprintf(file_lambda, lambda_format_r, sampled_lin{yy}(s, :));    
       end
       fprintf(file_lambda, '\n');    
   end   
   % Dump variances
   var_format = ' %.3f';
   for s = 2:endpoint
       for yy = 1:num_observed
           fprintf(file_yvar, var_format, sampled_yerr_par(s, yy));    
       end
       fprintf(file_yvar, '\n');    
   end   
end

end

