% mcmc_gpsem:
%
% This function generates MCMC samples from a GPSEM. 
%
% IT IS ASSUMED THAT ORDER OF OBSERVED VARIABLES IN THE GRAPH CORRESPONDS
% TO THE ORDER OF COLUMNS IN THE DATASET.
%
% Inputs:
%
% - n_mcmc_samples: number of MCMC samples to be generated.
% - graph: a valid SEM DAG;
% - n_types: a binary vector indicating which variables are observed
%            (== 1) and which are hidden (== 0)
% - lin_var: global prior variance hyperparameter for the linear
%            coefficients
% - K:       number of mixture components;
% - alpha:   weight for the Dirichlet prior of the mixture probability mass 
%            function;
% - hyper_x_mix: parameters for the Dirichlet mixture model.
% - hyper_y_var: inverse gamma hyperparameters for measurement variance.
% - fixed_loadings: cell array, one entry per observed variable. Each
%                   entry is a vector indicating which loadings are fixed
%                   to '1' (the size of the vector is the number of parents
%                   of y plus one, where the last entry corresponds to the
%                   intercept).
% - missing_values: a T x 2 matrix indicating missing values to be
%                   predicted. (t, 1) is the row of the t-th missing value
%                   and (t, 2) is the columno of the t-th missing value
%
% Prior for kernel hyperparameters will vary according to the kernel.

function [sampled_missing, sampled_x, ...
          sampled_lin, sampled_yerr_var, ...
          sampled_xmix_pi, sampled_xmix_mu, sampled_xmix_invcov, sampled_x_z] = ...
          mcmc_mog(num_mcmc_samples, K, graph, n_types, data, lin_var, ...
                   alpha_w, ...
                   hyper_x_mean_invcov, invcov_delta, invcov_U, hyper_y_var, ... 
                   fixed_loadings, fixed_loading_values, ...
                   missing_values)

mean_obs_data = mean(data); %TODO               

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PARSE GRAPH:
% x_ychildren: observed children of each latent. It contains the position
%              of each children wrt to the observed variables only, not 
%              the whole graph.
% x_xchildren: latent children of each latent. Again, numbers represent
%              position with respect the latent variables only.
% y_xparents:  latent parents of each observed variable (same indexing
%              comment applies).

num_nodes = size(graph, 1);
latents = find(n_types == 0);  observed = find(n_types == 1);
num_latents = length(latents); num_observed = length(observed);
children = cell(num_nodes, 1); parents = cell(num_nodes, 1);
for v = 1:num_nodes
    children{v} = find(graph(:, v) == 1)';
    parents{v} = find(graph(v, :) == 1);
end

x_ychildren = cell(num_latents, 1); 
x_xchildren = cell(num_latents, 1); x_xparents = cell(num_latents, 1);
exo_latent = zeros(num_latents, 1);
for x = 1:num_latents
    x_pos = latents(x);
    obs_children    = intersect(observed, children{x_pos});
    latent_children = intersect(latents, children{x_pos});
    latent_parents  = intersect(latents, parents{x_pos});
    x_ychildren{x} = zeros(1, length(obs_children));
    for v = 1:length(obs_children)
        x_ychildren{x}(v) = find(observed == obs_children(v), 1);
    end
    x_xchildren{x} = zeros(1, length(latent_children));
    for v = 1:length(latent_children)
        x_xchildren{x}(v) = find(latents == latent_children(v), 1);
    end
    x_xparents{x} = zeros(1, length(latent_parents));
    for v = 1:length(latent_parents)
        x_xparents{x}(v) = find(latents == latent_parents(v), 1);
    end
    exo_latent(x) = isempty(latent_parents);
end
y_xparents = cell(1, num_observed);
for y = 1:num_observed
    y_pos = observed(y);
    y_xparents{y} = zeros(1, length(parents{y_pos}));
    for v = 1:length(parents{y_pos})
        y_xparents{y}(v) = find(latents == parents{y_pos}(v), 1);
    end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ALLOCATE SPACE. Everything is being done in RAM now, so this can quickly
% exhauste the memory.

N  = size(data, 1);
sampled_x    = zeros(num_mcmc_samples, N, num_latents);
sampled_lin  = cell(num_observed, 1);
for v = 1:num_observed
    y = observed(v);
    sampled_lin{v} = zeros(num_mcmc_samples, length(parents{y}) + 1);
end
alpha = alpha_w * ones(K, 1);
sampled_xmix_pi     = zeros(num_mcmc_samples, K);
sampled_xmix_mu     = cell(num_mcmc_samples, K);
sampled_xmix_invcov = cell(num_mcmc_samples, K);
sampled_x_z         = zeros(num_mcmc_samples, N);
sampled_yerr_var    = zeros(num_mcmc_samples, num_observed);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% INITIAL SAMPLING. First, initialize error variances.
% After that, order latents in a partial order where a node 
% cannot come before any of its descendants. Then sample according to
% order.

sampled_yerr_var(1, :) = ones(num_observed, 1);
sampled_xmix_pi(1, :) = 1 / K;

for k = 1:K
   sampled_xmix_mu{1, k} = zeros(num_latents, 1);
   for x = 1:num_latents
       children_mean = mean(mean_obs_data(x_ychildren{x}));
       sampled_xmix_mu{1, k}(x) = children_mean + randn() * 0.01;
   end
   sampled_xmix_invcov{1, k} = eye(num_latents);
end

%sampled_x_z(1, :) = sample(sampled_xmix_pi(1, :), N); 
sampled_x_z(1, :) = ones(1, N);
for k = 1:K
    sel_z = sampled_x_z(1, :) == k;
    s = randnorm(sum(sel_z), sampled_xmix_mu{1, k}, [], inv(sampled_xmix_invcov{1, k}));
    sampled_x(1, sel_z, :) = s';
end

for y = 1:num_observed % And now, sample linear coefficients
    n_par = length(y_xparents{y});
    cov_y = eye(n_par + 1);
    sampled_lin{y}(1, :) = randnorm(1, zeros(n_par + 1, 1), [], cov_y)';
end

for y = 1:num_observed
    if sum(fixed_loadings{y}) == 0
       continue
    end
    n_par = length(y_xparents{y});
    if ~isempty(fixed_loading_values)
       sampled_lin{y}(1, fixed_loadings{y} == 1) = fixed_loading_values{y}';
    else
       sampled_lin{y}(1, fixed_loadings{y} == 1) = 1;
       if fixed_loadings{y}(n_par + 1)
          sampled_lin{y}(1, n_par + 1) = 0;
       end
    end
end

% Initialize missing values
num_missing = size(missing_values, 1);

if num_missing > 0
   sampled_missing = zeros(num_mcmc_samples, num_missing);
else
   sampled_missing = [];
end
for t = 1:num_missing
    r = missing_values(t, 1); c = missing_values(t, 2);
    data(r, c) = mean_obs_data(c) + 0.01 * randn();
    sampled_missing(1, t) = data(r, c);
end

file_x       = fopen('sample_x.txt','w');
file_y       = fopen('sample_y.txt','w');
file_lambda  = fopen('sample_lambda.txt','w');
file_yvar    = fopen('sample_yvar.txt','w');
file_z       = fopen('sample_z.txt','w');
file_pi      = fopen('sample_pi.txt','w');
file_invcov  = fopen('sample_invcov.txt','w');
file_mu      = fopen('sample_mu.txt','w');


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PERFORM SAMPLING

for iter = 2:num_mcmc_samples    

    disp(sprintf('Iteration %d', iter));
    
    % Sample latent variable marginal (mixture of Gaussians) parameters
    % First, sample means and variances for each mixture component        
    for k = 1:K
        sel_z  = sampled_x_z(iter - 1, :) == k;
        size_z = sum(sel_z);
        if size_z ~= 1
           s_x = squeeze(sum(sampled_x(iter - 1, sel_z, :)));
        else 
           % Stupid MATLAB behaves badly here
           s_x = zeros(num_latents, 1);
           z_idx = find(sel_z == 1, 1);
           for xx = 1:num_latents
               s_x(xx) = sampled_x(iter - 1, z_idx, xx);
           end
        end
           
        % Sample mean given inverse covariance variance matrix
        c_mean = inv(size_z * sampled_xmix_invcov{iter - 1, k} + ...
                     hyper_x_mean_invcov);        
        if size(c_mean, 2) ~= size(sampled_xmix_invcov{iter - 1, k}, 1) || ...
           size(sampled_xmix_invcov{iter - 1, k}, 2) ~= size(s_x, 1) || ...
           size(c_mean, 2) ~= size(s_x, 1)
           disp('!');
        end
        m_mean = c_mean * sampled_xmix_invcov{iter - 1, k} * s_x;
        new_mean = randnorm(1, m_mean, [], c_mean);
        sampled_xmix_mu{iter, k} = new_mean;
        % Sample inverse variance matrix given mean
        if size_z ~= 1
           centered_x = squeeze(sampled_x(iter - 1, sel_z, :)) - ...
                        repmat(new_mean', size_z, 1); 
        else
           % Correct another MATLAB oddity
           centered_x = squeeze(sampled_x(iter - 1, sel_z, :))' - ...
                        new_mean';
        end
        ss_x = centered_x' * centered_x;
        new_U = inv(ss_x + invcov_U);
        new_delta = invcov_delta + size_z;
        %sampled_xmix_invcov{iter, k} = wishrnd(new_U, new_delta);
        L = chol(new_U);
        W = randwishart(new_delta, num_latents);
        sampled_xmix_invcov{iter, k} = L' * W' * W * L;
    end
    % Now sample mixture component indicators given x and means
    weights = zeros(K, N);
    xd = squeeze(sampled_x(iter - 1, :, :))';
    for k = 1:K
        m  = sampled_xmix_mu{iter, k};
        v  = sampled_xmix_invcov{iter, k};       
        ld = logdet(v);
        for d = 1:N;
            cxd = xd(:, d) - m;
            weights(k, d) = 0.5 * ld - 0.5 * trace(v * cxd * cxd') + ...
                            log(sampled_xmix_pi(iter - 1, k));
        end
    end
    for d = 1:N
        weights_d = exp(weights(:, d) - max(weights(:, d)));
        weights_d = weights_d / sum(weights_d);
        sampled_x_z(iter, d) = sample(weights_d);
    end
    % Finally, sample mixture distribution
    alpha_pos = alpha;
    for k = 1:K
        alpha_pos(k)  = alpha_pos(k) + sum(sampled_x_z(iter, :) == k);
    end
    sampled_xmix_pi(iter, :) = randdirichlet(alpha_pos);
    
    % Sample measurement error variance hyperparameters
    for y = 1:num_observed
        if length(y_xparents{y}) > 1
           x_data = [squeeze(sampled_x(iter - 1, :, y_xparents{y})) repmat(1, N, 1)];
        else
           x_data = [sampled_x(iter - 1, :, y_xparents{y})' repmat(1, N, 1)]; 
        end
        ss_y = sum((data(:, y) - x_data * sampled_lin{y}(iter - 1, :)') .^ 2);
        ga = hyper_y_var(1) + N / 2;
        gb = hyper_y_var(2) + ss_y / 2;
        sampled_yerr_var(iter, y) = gb / randgamma(ga);
    end
    
    % Sample latent variables given mixture parameters, linear coefficients 
    % and data 
    % (TODO: ignore missing data instead of conditioning on samples)
    covs = cell(K, 1); marg_mean_cov = cell(K, 1);
    intercepts = zeros(num_observed, 1);
    B = zeros(num_observed, num_latents);
    V = diag(1 ./ sampled_yerr_var(iter, :));
    for y = 1:num_observed
        n_p = length(y_xparents{y});
        B(y, y_xparents{y}) = sampled_lin{y}(iter - 1, 1:n_p);
        intercepts(y) = sampled_lin{y}(iter - 1, n_p + 1);
    end
    for k = 1:K
        covs{k} = inv(B' * V * B + sampled_xmix_invcov{iter, k});
        marg_mean_cov{k} = sampled_xmix_invcov{iter, k} * sampled_xmix_mu{iter, k};
        sel_z = sampled_x_z(iter, :) == k;
        sampled_x(iter, sel_z, :) = randnorm(sum(sel_z), zeros(num_latents, 1), [], covs{k})';
    end
    for d = 1:N
        y_centered = data(d, :)' - intercepts;
        k          = sampled_x_z(iter, d);        
        mean_x     = covs{k} * (B' * V * y_centered + marg_mean_cov{k});
        sampled_x(iter, d, :) = squeeze(sampled_x(iter, d, :)) + mean_x;
    end
    
    % Sample linear coefficients given latent variables and data
    for y = 1:num_observed        
        n_par = length(y_xparents{y});
        if length(y_xparents{y}) > 1
           x_data = [squeeze(sampled_x(iter, :, y_xparents{y})) repmat(1, N, 1)];
        else
           x_data = [sampled_x(iter, :, y_xparents{y})' repmat(1, N, 1)];
        end
        cov_y = x_data' * x_data / sampled_yerr_var(iter, y);
        for i = 1:(n_par + 1)
            cov_y(i, i) = cov_y(i, i) + 1 / lin_var;
        end
        cov_y = inv(cov_y);
        mean_y = cov_y * sum(repmat(data(:, y), 1, n_par + 1) .* x_data)' / ...
                         sampled_yerr_var(iter, y); 
        if sum(fixed_loadings{y} == 0)
            sampled_lin{y}(iter, :) = randnorm(1, mean_y, [], cov_y)';        
        else
            fr = fixed_loadings{y} == 0; % Free coefficients   TODO
            fx = fixed_loadings{y} == 1; % Fixed coefficients
            if ~isempty(fixed_loading_values)
               sampled_lin{y}(iter, fx) = fixed_loading_values{y}';
            else
               sampled_lin{y}(iter, fx) = 1;            
               if fixed_loadings{y}(n_par + 1)
                  sampled_lin{y}(iter, n_par + 1) = 0;
               end
            end            
            reg_coeff = cov_y(fr, fx) * inv(cov_y(fx, fx));
            mean_f = reg_coeff * (sampled_lin{y}(iter, fx)' - mean_y(fx));
            cov_f  = cov_y(fr, fr) - reg_coeff * cov_y(fx, fr);
            sampled_lin{y}(iter, fr) = randnorm(1, mean_f, [], cov_f)';
        end
    end
    
    % Sample missing values
    if num_missing > 0
       sds = sqrt(sampled_yerr_var(iter, :));
       rand_pool = randn(num_missing, 1);
       for t = 1:num_missing
           d = missing_values(t, 1);
           y = missing_values(t, 2);        
           m_t = sampled_lin{y}(iter, :) * [sampled_x(iter, d, y_xparents{y}); 1];
           data(d, y) = m_t + sds(y) * rand_pool(t);
           sampled_missing(iter, t) = data(d, y);
       end
    end
end

%print_to_file(num_mcmc_samples);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Print all points in the batch to the respective files

function print_to_file(endpoint)
   disp('Printing to file...');
   % Dump latent variables
   x_format = ' %.3f';
   x_format_r = x_format;
   for w = 2:num_latents
       x_format_r = strcat(x_format_r, x_format);
   end
   x_format_r = strcat(x_format_r, '\n');
   for s = 2:endpoint
       fprintf(file_x, x_format_r, sampled_x(s, :, :));    
   end
   % Dump imputed data
   if num_missing > 0
      y_format = ' %.3f';
      y_format_r = y_format;
      for w = 1:(num_missing - 1)
          y_format_r = strcat(y_format_r, y_format);
      end
      y_format_r = strcat(y_format_r, '\n');
      for s = 2:endpoint
          fprintf(file_y, y_format_r, sampled_missing(s, :));    
      end
   end
   % Dump mixture indicators
   z_format = ' %d';
   z_format_r = z_format;
   for w = 2:N
       z_format_r = strcat(z_format_r, z_format);
   end
   z_format_r = strcat(z_format_r, '\n');
   for s = 2:endpoint
       fprintf(file_z, z_format_r, sampled_x_z(s, :));    
   end
   % Dump mixture information
   pi_format = ' %.3f';
   pi_format_r = pi_format;
   for w = 2:K
       pi_format_r = strcat(pi_format_r, pi_format);
   end
   pi_format_r = strcat(pi_format_r, '\n');
   for s = 2:endpoint
       fprintf(file_pi, pi_format_r, sampled_xmix_pi(s, :));
   end
   invcov_format = ' %.10f';
   invcov_format_r = invcov_format;
   for i = 2:num_latents^2
       invcov_format_r = strcat(invcov_format_r, invcov_format);
   end
   invcov_format_r = strcat(invcov_format_r, '\n');
   for s = 2:endpoint
       for k = 1:K
           fprintf(file_invcov, invcov_format_r, sampled_xmix_invcov{s, k});
       end
   end   
   mu_format = ' %.3f';
   mu_format_r = mu_format;
   for w = 2:num_latents
        mu_format_r = strcat(mu_format_r, mu_format);
   end
   mu_format_r = strcat(mu_format_r, '\n');
   for s = 2:endpoint
       for k = 1:K
           fprintf(file_mu, mu_format_r, sampled_xmix_mu{s, k});
       end
   end
   % Dump linear coefficient hyperparameters
   lambda_format = ' %.3f';
   for s = 2:endpoint
       for y = 1:num_observed
           lambda_format_r = lambda_format;
           for w = 1:length(y_xparents{y})
               lambda_format_r = strcat(lambda_format_r, lambda_format);
           end
           fprintf(file_lambda, lambda_format_r, sampled_lin{y}(s, :));    
       end
       fprintf(file_lambda, '\n');    
   end   
   % Dump variances
   var_format = ' %.3f';
   for s = 2:endpoint
       for y = 1:num_observed
           fprintf(file_yvar, var_format, sampled_yerr_var(s, y));    
       end
       fprintf(file_yvar, '\n');    
   end   
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

fclose(file_x);
fclose(file_y);
fclose(file_lambda);
fclose(file_yvar);
fclose(file_z);
fclose(file_pi);
fclose(file_invcov);
fclose(file_mu);

end

