%%%%%%%%%%%%
% pred_llik_finite:
%
% Same as joint_llik, but for the finite-basis model only.

function [llik y_fake x_fake] = pred_llik_finite(graph, K, num_mcmc, num_latents, test_rows, data, dir_name, burn_in, NUM_FOLDS)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% INITIALIZATION

%%% Get some graph information
num_y        = size(graph, 1) - num_latents;
exo_x        = get_exo_vars(graph(1:num_latents, 1:num_latents));
endo_x       = setdiff(1:num_latents, exo_x);
num_exo      = length(exo_x);
latent_order = get_node_order(graph(1:num_latents, 1:num_latents));
latent_order = setdiff(latent_order, exo_x)';
x_parents    = graph((num_latents + 1):size(graph, 1), 1:num_latents);
num_parents  = sum(x_parents, 2);
x_xparents   = cell(num_latents, 1);
for i = 1:num_latents
    x_xparents{i} = find(graph(i, :) == 1);
end
xp_list = cell(num_y, 1);
for y = 1:num_y
    xp_list{y} = [find(x_parents(y, :) == 1) (num_latents + 1)];
end
num_lambda  = sum(sum(x_parents)) + num_y; % Total number of parents, plus intercepts
weights     = cell(num_latents, 1);
weight_size = zeros(num_latents, 1);
for x = 1:num_latents
    if isempty(x_xparents{x})
       continue
    end
    weight_size(x) = basisfunc(length(x_xparents{x}));
    weights{x} = zeros(num_mcmc, weight_size(x));    
end

%%% Allocate some buffer information
lambda      = zeros(num_y, num_latents + 1);
llik        = cell(NUM_FOLDS, 1);
y_fake      = zeros(num_mcmc - burn_in, num_y);
x_fake      = zeros(num_mcmc - burn_in, num_latents);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% ITERATION

%%% Now, iterate through folds, and within folds iterate through MCMC
%%% samples
for k = 1:NUM_FOLDS
   
    %%% Get some data information
    test_size   = length(test_rows{k});

    %%% Allocate more buffer information
    mu_y        = zeros(test_size, num_y);
    llik{k}     = zeros(test_size, 1);
    llik_k      = zeros(num_mcmc - burn_in, test_size);
    
    %%% Open files
    file_id_w   = fopen(strcat(dir_name, '\sample_w.', int2str(k), '.log'), 'r');
    file_id_xv  = fopen(strcat(dir_name, '\sample_xvar.', int2str(k), '.log'), 'r');    
    file_id_exo = fopen(strcat(dir_name, '\sample_exo.', int2str(k), '.log'), 'r');
    file_id_l   = fopen(strcat(dir_name, '\sample_lambda.', int2str(k), '.log'), 'r');
    file_id_v   = fopen(strcat(dir_name, '\sample_yvar.', int2str(k), '.log'), 'r');    

	w_line      = textscan(file_id_w,   '%f', sum(weight_size));
    varx_line   = textscan(file_id_xv,  '%f', num_latents);
    exopar_line = textscan(file_id_exo, '%f', num_exo * K * 3);
    lambda_line = textscan(file_id_l,   '%f', num_lambda);
    var_line    = textscan(file_id_v,   '%f', num_y);

    %%% Iterate through the MCMC samples
    mcmc_count  = 1;    
    while ~feof(file_id_v) && mcmc_count <= num_mcmc 
        disp(sprintf('folder %d, mcmc #%d', k, mcmc_count));
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % START Read sampled training latent variables, and parameters
        
        %----- Read weights
        buffer_w = w_line{1}; buffer_pos = 1;
        for x = endo_x
            weights{x} = buffer_w(buffer_pos:(buffer_pos + weight_size(x) - 1));
            buffer_pos = buffer_pos + weight_size(x);
        end
        w_line = textscan(file_id_w,  '%f', sum(weight_size));
        %----- Read latent conditional variances
        vx = varx_line{1}';
        varx_line   = textscan(file_id_xv,  '%f', num_latents);
        %----- Read exogenous mixture parameters
        exo_params  = reshape(exopar_line{1}, K * 3, num_exo)';
        exopar_line = textscan(file_id_exo, '%f', num_exo * K * 3);
        %----- Read measurement model
        p_count = 1;
        for y = 1:num_y
            lambda(y, xp_list{y}) = lambda_line{1}(p_count:(p_count + num_parents(y)))';
            p_count = p_count + 1 + num_parents(y);
        end 
        lambda_line = textscan(file_id_l,   '%f', num_lambda);
        v = var_line{1}';
        var_line    = textscan(file_id_v,   '%f', num_y);
        
        % END Read sampled training latent variables, and parameters
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %%% Decides whether to count this or not
        if mcmc_count <= burn_in
           mcmc_count = mcmc_count + 1;
           continue;
        end
        
        %%% Now, sample new latent variables
        xstar = zeros(test_size, num_latents + 1); 
        % Exogenous variables first
        for x_pos = 1:num_exo
            prob = zeros(K, 1);
            for kk = 1:K
                prob(kk) = exo_params(x_pos, (kk - 1) * 3 + 1);
            end
            z = sample(prob, test_size)';
            mean_x = exo_params(x_pos, (z - 1) * 3 + 2)';
            var_x  = exo_params(x_pos, (z - 1) * 3 + 3)';
            xstar(:, exo_x(x_pos)) = randn(test_size, 1) .* sqrt(var_x) + mean_x;
        end
        % Endogenous variables (following partial order)
        for x = latent_order
            parent_data = basisfunc(xstar, x_xparents{x});
            mean_x = parent_data * weights{x};
            xstar(:, x) = randn(test_size, 1) * sqrt(vx(x)) + mean_x;
        end
        % Add intercept term
        xstar(:, num_latents + 1) = ones(test_size, 1);        
        
        %%% Calculate the loglikelihood of Y* given X* and store it
        for d = 1:test_size
            mu_y(d, :) = sum(lambda .* repmat(xstar(d, :), num_y, 1), 2)';
        end
        v_matrix = repmat(v, test_size, 1);
        llik_k(mcmc_count - burn_in, :) = sum(-0.5 * log(v_matrix) ...
                                          -0.5 * ((data{k}(test_rows{k}, :) - mu_y) .^ 2) ./ v_matrix, 2);
                                 
        %%% Generate fake data
        y_fake(mcmc_count - burn_in, :) = randn(1, num_y) .* sqrt(v) + mu_y(1, :);
        x_fake(mcmc_count - burn_in, :) = xstar(1, 1:num_latents);

        %%% End of MCMC iteration                              
        mcmc_count = mcmc_count + 1;
    end
    
    %%% End of folder iteration
    
    llik_k = llik_k(1:(mcmc_count - burn_in - 1), :);
    num_llik_k = size(llik_k, 1);
    for d = 1:test_size
        m = max(llik_k(:, d));
        llik{k}(d) = log(sum(exp(llik_k(:, d) - m))) + m...
                     -0.5 * num_y * log(2 * pi) -log(num_llik_k);
    end
    
    fclose(file_id_w);    fclose(file_id_xv);
    fclose(file_id_exo);  fclose(file_id_l);     fclose(file_id_v);

end



% Example code:
%
% all_data = cell(NUM_FOLDS, 1); for i = 1:NUM_FOLDS, all_data{i} = normalize_training(raw_data, all_train_points{i}, zeros(size(raw_data))); end
% burn_in = 2500; dir_name = '.'; num_mcmc_samples = 20000;
% [llik y_fake x_fake] = pred_llik_finite(graph, num_x_mix, num_mcmc_samples, num_latents, all_test_points, all_data, dir_name, burn_in, NUM_FOLDS);
